from OFS.SimpleItem import SimpleItem

from zope.interface import implements, Interface
from zope.component import adapts
from zope.formlib import form

from zope.component.interfaces import IObjectEvent

from plone.contentrules.rule.interfaces import IExecutable, IRuleElementData

from plone.app.contentrules.browser.formhelper import AddForm, EditForm

from Acquisition import aq_inner

from seantis.dir.events import _
from seantis.dir.events.interfaces import ISourceCondition


class SourceCondition(SimpleItem):

    """The actual persistent implementation of the source condition element.
    
    Note that we must mix in SimpleItem to keep Zope 2 security happy.
    """
    implements(ISourceCondition, IRuleElementData)

    source = ''
    element = 'seantis.dir.events.conditions.source'

    @property
    def summary(self):
        return _(u"Source contains: ${source}",
                 mapping={'source': self.source})


class SourceConditionExecutor(object):

    """The executor for this condition.
    
    This is registered as an adapter in configure.zcml
    """
    implements(IExecutable)
    adapts(Interface, ISourceCondition, IObjectEvent)

    def __init__(self, context, element, event):
        self.context = context
        self.element = element
        self.event = event

    def __call__(self):
        context = aq_inner(self.event.object)
        try:
            source = context.source
        except (AttributeError, TypeError,):
            # The object doesn't have a source
            return False
        if not self.element.source:
            # Any source
            return True
        try:
            matched = (source.split('/')[-1] == self.element.source)
            return matched
        except:
            return False


class SourceConditionAddForm(AddForm):

    """An add form for portal type conditions.
    """
    form_fields = form.FormFields(ISourceCondition)
    label = _(u"Add import source condition")
    description = _(
        u"An import source condition makes the rule apply only to content "
        u"with the given import source.")
    form_name = _(u"Configure source condition")

    def create(self, data):
        c = SourceCondition()
        form.applyChanges(c, self.form_fields, data)
        return c


class SourceConditionEditForm(EditForm):

    """An edit form for portal type conditions
    """
    form_fields = form.FormFields(ISourceCondition)
    label = _(u"Edit import source condition")
    description = _(
        u"An import source condition makes the rule apply only to content "
        u"with the given import source.")
    form_name = _(u"Configure source condition")



import string

from logging import getLogger
log = getLogger('seantis.dir.events')

from datetime import timedelta
from dateutil.parser import parse

from lxml import objectify
from urllib2 import urlopen

from five import grok
from zope.component import queryAdapter

from seantis.dir.events.interfaces import (
    IExternalEventCollector,
    IExternalEventSourceGuidle,
    IGuidleClassifier,
    NoImportDataException
)


class EventsSourceGuidle(grok.Adapter):
    grok.context(IExternalEventSourceGuidle)
    grok.provides(IExternalEventCollector)

    def generate_recurrence(self, date):
        try:
            weekdays = list(date.weekdays.iterchildren())
            return "RRULE:FREQ=WEEKLY;BYDAY=%s" % ','.join(
                [d.text.upper() for d in weekdays]
            )
        except AttributeError:
            return ""

    def limit_recurrence(self, recurrence, end):
        if not end or not recurrence:
            return recurrence
        else:
            return recurrence + ';UNTIL=%s' % end.strftime('%Y%m%dT%H%MZ')

    def get_dates(self, date):

        def child(name):
            try:
                return getattr(date, name).text
            except AttributeError:
                return None

        return (
            child('startDate'), child('endDate'),
            child('startTime'), child('endTime')
        )

    def events(self, offer):
        """ Generates one event for each date in the guidle input. """

        dates = [d for d in offer.schedules.iterchildren()]
        if not dates:
            log.error(
                'offer with id %s has no dates, skipping' % offer.attrib['id']
            )

            raise StopIteration

        for date in dates:

            start, end, start_time, end_time = self.get_dates(date)

            event = {}

            event['id'] = offer.attrib['id']
            event['recurrence'] = self.generate_recurrence(date)
            event['start'] = parse(start)

            # the end date of recurring events seems to be the date of the
            # last occurrence, which makes sense since guidle only seems
            # to support daily occurences by weekday
            if end and not event['recurrence']:
                event['end'] = parse(end)
            else:
                # if a recurrence exists it needs to be limited by the end date
                if event['recurrence']:
                    until = parse(end) + timedelta(days=1)
                    event['recurrence'] = self.limit_recurrence(
                        event['recurrence'], until
                    )

                event['end'] = event['start']

            if not any((start_time, end_time)):
                event['whole_day'] = True
            else:
                event['whole_day'] = False

            if start_time:
                event['start'] = parse(start_time, default=event['start'])

            if end_time:
                event['end'] = parse(end_time, default=event['end'])

            # if the event ends before it starts it means that we have a range
            # like this: 20:00 - 00:30, so the event really ends the next day
            if event['end'] < event['start']:
                event['end'] += timedelta(days=1)

            yield event

    def copy(self, event, node, expression):
        lines = map(string.strip, expression.split('\n'))

        for line in lines:
            if not line:
                continue

            key, children = map(string.strip, line.split('<-'))

            for child in map(string.strip, children.split(',')):
                if not hasattr(node, child):
                    continue

                if key in event:
                    event[key] += '\n\n' + getattr(node, child).text
                else:
                    event[key] = getattr(node, child).text

    def fetch(self, xml=None):
        try:
            if xml is None:
                xml = urlopen(self.context.url, timeout=60).read()
            root = objectify.fromstring(xml)

            offers = root.xpath(
                '*//guidle:offer',
                namespaces={'guidle': 'http://www.guidle.com'}
            )
        except:
            raise NoImportDataException()

        classifier = queryAdapter(self, IGuidleClassifier)

        for offer in offers:

            last_update_of_offer = parse(
                offer.lastUpdateDate.text
            ).replace(
                microsecond=0
            )

            assert last_update_of_offer, """"
                offers should all have an update timestamp
            """

            for e in self.events(offer):

                e['fetch_id'] = self.context.url
                e['last_update'] = last_update_of_offer

                # so far all guidle events seem to be in this region
                e['timezone'] = 'Europe/Zurich'
                e['source_id'] = offer.attrib['id']

                # basic information
                self.copy(e, offer.offerDetail, """
                    title               <- title
                    short_description   <- shortDescription
                    long_description    <- longDescription, openingHours
                    prices              <- priceInformation
                    event_url           <- externalLink
                    location_url        <- homepage
                    registration        <- ticketingUrl
                """)

                # address
                self.copy(e, offer.address, """
                    locality    <- company
                    street      <- street
                    zipcode     <- zip
                    town        <- city
                    latitude    <- latitude
                    longitude   <- longitude
                """)

                # contact
                self.copy(e, offer.contact, """
                    organizer       <- company
                    contact_name    <- name
                    contact_email   <- email
                    contact_phone   <- telephone_1
                """)

                # categories
                e['cat1'] = set()
                if classifier:
                    e['cat1'] = classifier.classify(offer.classifications)
                e['cat2'] = set((e['town'],))

                # image (download later)
                try:
                    for image in list(
                            offer.offerDetail.images.iterchildren())[:1]:
                        self.copy(e, image, "image <- url")
                except AttributeError:
                    pass

                # attachments (download later)
                try:
                    attachments = list(
                        offer.offerDetail.attachments.iterchildren()
                    )[:2]
                    for ix, attachment in enumerate(attachments):
                        self.copy(e, attachment,
                                  "attachment_%i <- url" % (ix + 1))
                except AttributeError:
                    pass

                yield e



import logging
log = logging.getLogger('seantis.dir.events')

from five import grok

from datetime import date
from dateutil.parser import parse

from zope.component import queryAdapter
from zope.interface import implements
from zope.event import notify
from Products.CMFPlone.PloneBatch import Batch

from seantis.dir.base import directory
from seantis.dir.base import session
from seantis.dir.base.utils import cached_property

from seantis.dir.events.interfaces import (
    IEventsDirectory, IActionGuard, IResourceViewedEvent, IExternalEvent
)

from seantis.dir.events.recurrence import grouped_occurrences
from seantis.dir.events import dates
from seantis.dir.events import utils
from seantis.dir.events import _

from AccessControl import getSecurityManager
from Products.CMFCore import permissions


class ResourceViewedEvent(object):
    implements(IResourceViewedEvent)

    def __init__(self, context):
        self.context = context


class EventsDirectory(directory.Directory):

    def labels(self):
        return dict(cat1=_(u'What'), cat2=_(u'Where'))

    def used_categories(self):
        return ('cat1', 'cat2')

    def unused_categories(self):
        return ('cat3', 'cat4')

    def allow_action(self, action, item_brain):
        """ Return true if the given action is allowed. This is not a
        wrapper for the transition guards of the event workflow. Instead
        it is called *by* the transition guards.

        This allows a number of people to work together on an event website
        with every person having its own group of events which he or she is
        responsible for.

        Per default, only external events may be hidden and only internal
        events archived.

        A client specific packages like izug.seantis.dir.events may
        use a custom adapter to override the default behaviour.
        """
        result = True

        try:
            IExternalEvent(item_brain)
            external_event = True
        except:
            external_event = False

        if external_event:
            if action not in ['hide', 'publish']:
                result = False
        else:
            if action == 'hide':
                result = False

        guard = queryAdapter(self, IActionGuard)
        if guard:
            result = guard.allow_action(action, item_brain)
        return result


class ExtendedDirectoryViewlet(grok.Viewlet):
    grok.context(IEventsDirectory)
    grok.name('seantis.dir.events.directory.detail')
    grok.require('zope2.View')
    grok.viewletmanager(directory.DirectoryViewletManager)

    template = grok.PageTemplateFile('templates/directorydetail.pt')

    def __init__(self, *args, **kwargs):
        super(ExtendedDirectoryViewlet, self).__init__(*args, **kwargs)


class EventsDirectoryIndexView(grok.View, directory.DirectoryCatalogMixin):

    grok.name('eventindex')
    grok.context(IEventsDirectory)
    grok.require('cmf.ManagePortal')

    template = None

    def render(self):

        self.request.response.setHeader("Content-type", "text/plain")

        if 'rebuild' in self.request:
            log.info('rebuilding ZCatalog')
            self.catalog.catalog.clearFindAndRebuild()

        if 'reindex' in self.request:
            log.info('reindexing event indices')
            self.catalog.reindex()

        result = []
        for name, index in self.catalog.indices.items():

            result.append(name)
            result.append('-' * len(name))
            result.append('')

            for ix, identity in enumerate(index.index):
                result.append('%i -> %s' % (ix, identity))

            result.append('')

            dateindex = index.get_metadata('dateindex')

            if dateindex:
                result.append('-> dateindex')

                for date in sorted(dateindex):
                    result.append('%s -> %s' % (
                        date.strftime('%y.%m.%d'), dateindex[date])
                    )

        return '\n'.join(result)


class EventsDirectoryView(directory.View):

    grok.name('view')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    template = None
    _template = grok.PageTemplateFile('templates/directory.pt')

    fired_event = False

    @property
    def title(self):
        return self.directory.title

    @property
    def is_ical_export(self):
        """ Returns true if the current request is an ical request. """
        return self.request.get('type') == 'ical'

    @property
    def is_json_export(self):
        """ Returns true if the current request is an json request. """
        return self.request.get('type') == 'json'

    def get_last_daterange(self):
        """ Returns the last selected daterange. """
        return session.get_session(self.context, 'daterange') \
            or dates.default_daterange

    def set_last_daterange(self, method):
        """ Store the last selected daterange on the session. """
        session.set_session(self.context, 'daterange', method)

    def get_last_custom_start_date(self):
        """ Returns the last selected custom start date. """
        default, unused = getattr(dates.DateRanges(), 'custom')
        return session.get_session(self.context,
                                   'custom_date_start') or default

    def get_last_custom_end_date(self):
        """ Returns the last selected custom end date. """
        unused, default = getattr(dates.DateRanges(), 'custom')
        return session.get_session(self.context,
                                   'custom_date_end') or default

    def set_last_custom_dates(self, start, end):
        """ Store the last selected custom dates on the session. """
        session.set_session(self.context, 'custom_date_start', start)
        session.set_session(self.context, 'custom_date_end', end)

    def get_last_state(self):
        """ Returns the last selected event state. """
        return session.get_session(self.context, 'state') or 'published'

    def set_last_state(self, method):
        """ Store the last selected event state on the session. """
        session.set_session(self.context, 'state', method)

    def get_last_import_source(self):
        """ Returns the last selected import source. """
        return session.get_session(self.context, 'source') \
            or ''

    def set_last_import_source(self, source):
        """ Store the last selected import source on the session. """
        session.set_session(self.context, 'source', source)

    @property
    def no_events_helptext(self):
        if 'published' == self.catalog.state:
            return _(u'No events for the current daterange')
        else:
            return _(u'No events for the current state')

    @property
    def selected_daterange(self):
        return self.catalog.daterange

    @property
    def dateranges(self):
        return dates.methods

    def daterange_url(self, method):
        return self.directory.absolute_url() + '?range=' + method

    @property
    def custom_date_min(self):
        return date.today().strftime('%Y-%m-%d')

    @property
    def custom_date_max(self):
        mim_date, max_date = dates.eventrange()
        return max_date.strftime('%Y-%m-%d')

    @property
    def custom_date_from(self):
        return self.catalog.custom_start_date().strftime('%Y-%m-%d')

    @property
    def custom_date_to(self):
        return self.catalog.custom_end_date().strftime('%Y-%m-%d')

    @property
    def custom_date_url(self):
        url = self.daterange_url('custom')
        url += '&from=' + self.custom_date_from
        url += '&to=' + self.custom_date_to
        return url

    @cached_property
    def locale(self):
        # borrowed from widget collective.z3form.datetimewidget.DateWidget
        cal = self.request.locale.dates.calendars['gregorian']
        locale = {
            'lang': getattr(self.request, 'LANGUAGE', 'en'),
            'months': ','.join(cal.getMonthNames()),
            'shortmonths': ','.join(cal.getMonthAbbreviations()),
            'days': ','.join([cal.getDayNames()[6]] + cal.getDayNames()[:6]),
            'shortdays': ','.join(
                [cal.getDayAbbreviations()[6]] + cal.getDayAbbreviations()[:6]
            ),
            'format': 'dd.mm.yyyy'
        }
        return locale

    @property
    def has_results(self):
        return len(self.batch) > 0

    @property
    def show_import_sources(self):
        return getSecurityManager().checkPermission(
            permissions.ReviewPortalContent, self.context
        )

    @property
    def import_sources(self):
        sources = [(source.Title, source.id) for
                   source in self.catalog.import_sources()]
        sources.insert(0, (u'-', ''))
        return sources

    def import_source_url(self, source):
        return self.directory.absolute_url() + '?source=' + source

    @property
    def selected_import_source(self):
        return self.catalog.import_source

    @property
    def import_sources_config(self):
        return [(source.Title, source.getURL())
                for source in self.catalog.import_sources()]

    def render(self):
        """ Renders the ical/json if asked, or the usual template. """
        search = 'search' in self.request.keys() \
            and self.request.get('searchtext') or None
        term = 'filter' in self.request.keys() \
            and self.get_filter_terms() or None
        max = int(self.request.get('max', 0))
        compact = 'compact' in self.request.keys()

        if self.is_ical_export:
            calendar = self.catalog.calendar(search=search, term=term)
            return utils.render_ical_response(self.request, self.context,
                                              calendar)

        elif self.is_json_export:
            export = self.catalog.export(search=search, term=term, max=max)
            return utils.render_json_response(self.request, export, compact)

        else:
            return self._template.render(self)

    def update(self, **kwargs):
        # Update date range
        daterange = self.request.get('range', self.get_last_daterange())
        if not dates.is_valid_daterange(daterange):
            daterange = 'this_month'
        else:
            self.set_last_daterange(daterange)

        # Update custom dates
        start = self.get_last_custom_start_date()
        end = self.get_last_custom_end_date()
        if daterange == 'custom':
            tz = start.tzinfo
            try:
                start = parse(self.request.get('from'), ignoretz=True)
                start = start.replace(tzinfo=tz)
                end = parse(self.request.get('to'), ignoretz=True)
                end = end.replace(tzinfo=tz)
                end = max(start, end)
                self.set_last_custom_dates(start, end)
            except:
                pass

        # Update state
        state = self.request.get('state', self.get_last_state())
        if not self.show_state_filters or state not in (
            'submitted', 'published', 'archived', 'hidden'
        ):
            state = 'published'
        else:
            self.set_last_state(state)

        # Update import source
        source = self.request.get('source', self.get_last_import_source())
        if not self.show_import_sources:
            source = ''
        else:
            self.set_last_import_source(source)

        self.catalog.state = state
        self.catalog.daterange = daterange
        self.catalog.import_source = source
        self.catalog.set_custom_dates(start, end)

        if not self.is_ical_export and not self.is_json_export:
            super(EventsDirectoryView, self).update(**kwargs)

        if not self.fired_event:
            notify(ResourceViewedEvent(self.context))
            self.fired_event = True

    def groups(self, items):
        """ Returns the given occurrences grouped by human_date. """
        groups = grouped_occurrences(items, self.request)

        for key, items in groups.items():
            for ix, item in enumerate(items):
                items[ix] = item.get_object()

        return groups

    def translate(self, text, domain="seantis.dir.events"):
        return utils.translate(self.request, text, domain)

    @property
    def show_state_filters(self):
        return getSecurityManager().checkPermission(
            permissions.ReviewPortalContent, self.context
        )

    @cached_property
    def batch(self):
        # use a custom batch whose items are lazy evaluated on __getitem__
        start = int(self.request.get('b_start') or 0)
        lazy_list = self.catalog.lazy_list

        # seantis.dir.events lazy list implementation currently cannot
        # deal with orphans.
        return Batch(lazy_list, directory.ITEMSPERPAGE, start, orphan=0)

    @property
    def selected_state(self):
        return self.catalog.state

    def state_filter_list(self):

        submitted = utils.translate(self.request, _(u'Submitted'))
        submitted += u' (%i)' % self.catalog.submitted_count

        return [
            ('submitted', submitted),
            ('published', _(u'Published')),
            ('hidden', _(u'Hidden'))
        ]

    def state_url(self, method):
        return self.directory.absolute_url() + '?state=' + method

    @utils.webcal
    def ical_url(self, for_all):
        """ Returns the ical url of the current view. """
        url = self.daterange_url('this_year') + '&type=ical'

        if for_all:
            return url

        action, param = self.primary_action()

        if action not in (self.search, self.filter):
            return ''

        if action == self.search:
            if param:
                return url + '&search=true&searchtext=%s' % param
            else:
                return ''

        if action == self.filter:
            terms = dict([(k, v) for k, v in param.items() if v != '!empty'])

            if not terms:
                return ''

            url += '&filter=true'

            for item in terms.items():
                url += '&%s=%s' % item

            return url


class TermsView(grok.View):

    grok.name('terms')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    label = _(u'Terms and Conditions')
    template = grok.PageTemplateFile('templates/terms.pt')



import transaction

from datetime import date, datetime
from seantis.dir.events import dates
from seantis.dir.events.tests import IntegrationTestCase

from seantis.dir.events.catalog import (
    LazyList,
    EventOrderIndex,
    attach_reindex_to_transaction,
    ReindexDataManager,
)

from mock import Mock


class TestEventIndex(IntegrationTestCase):

    def test_lazy_list(self):

        # test indexing
        lazy = LazyList(Mock(), 10)
        self.assertEqual(len(lazy), 10)
        self.assertEqual(lazy._get_item.call_count, 0)

        lazy[0]
        self.assertEqual(lazy._get_item.call_count, 1)

        lazy[1]
        self.assertEqual(lazy._get_item.call_count, 2)

        self.assertRaises(IndexError, lambda: lazy[10])

        # test iterating
        lazy = LazyList(Mock(), 10)
        [i for i in lazy]
        self.assertEqual(lazy._get_item.call_count, 10)

        # test_slicing
        lazy = LazyList(Mock(), 10)
        lazy[:2]
        self.assertEqual(lazy._get_item.call_count, 2)

        lazy = LazyList(Mock(), 10)
        lazy[5:100]
        self.assertEqual(lazy._get_item.call_count, 5)

    def test_event_order_index(self):

        self.login_testuser()

        submitted = self.catalog.indices['submitted']
        published = self.catalog.indices['published']

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        event = self.create_event()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        event.submit()
        transaction.commit()

        self.assertEqual(len(submitted.index), 1)
        self.assertEqual(len(published.index), 0)

        event.publish()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 1)

        event.archive()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        event.publish()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 1)

        denied = self.create_event()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 1)

        denied.submit()
        transaction.commit()

        self.assertEqual(len(submitted.index), 1)
        self.assertEqual(len(published.index), 1)

        denied.do_action("deny")
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 1)

    def test_event_order_index_more(self):

        self.login_testuser()

        submitted = self.catalog.indices['submitted']
        published = self.catalog.indices['published']

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        # Lazy-create events on access
        events = LazyList(lambda i: self.create_event(), 50)

        num_submitted = 0
        num_published = 0

        for new_idx, new_event in enumerate(events):

            if new_idx > 0:

                for old_idx, old_event in enumerate(events[:new_idx - 1]):

                    if old_event.state == 'preview':

                        if (old_idx % 13 != 0):
                            old_event.submit()
                            transaction.commit()
                            num_submitted += 1

                    elif old_event.state == 'submitted':

                        if (old_idx % 7 != 0):
                            old_event.publish()
                            transaction.commit()
                            num_submitted -= 1
                            num_published += 1
                        else:
                            old_event.do_action("deny")
                            transaction.commit()
                            num_submitted -= 1

                    elif old_event.state == 'published':

                        if (old_idx % 5 != 0):
                            old_event.archive()
                            transaction.commit()
                            num_published -= 1

                    elif old_event.state == 'archived':

                        if (old_idx % 11 == 0):
                            old_event.publish()
                            transaction.commit()
                            num_published += 1

                    self.assertEqual(len(submitted.index), num_submitted)
                    self.assertEqual(len(published.index), num_published)

    def test_event_order_index_recurrence(self):
        self.login_testuser()

        submitted = self.catalog.indices['submitted']
        published = self.catalog.indices['published']

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        event = self.create_event(recurrence='RRULE:FREQ=DAILY;COUNT=10')
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        event.submit()
        transaction.commit()

        self.assertEqual(len(submitted.index), 10)
        self.assertEqual(len(published.index), 0)

        event.publish()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 10)

        event.archive()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        event.publish()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 10)

        event.recurrence = ''
        transaction.commit()

        event.reindexObject()
        event.reindex()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 1)

        event.archive()
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        denied = self.create_event(recurrence='RRULE:FREQ=DAILY;COUNT=10')
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        denied.submit()
        transaction.commit()

        self.assertEqual(len(submitted.index), 10)
        self.assertEqual(len(published.index), 0)

        denied.do_action("deny")
        transaction.commit()

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

    def test_event_order_index_recurrence_more(self):

        self.login_testuser()

        submitted = self.catalog.indices['submitted']
        published = self.catalog.indices['published']

        self.assertEqual(len(submitted.index), 0)
        self.assertEqual(len(published.index), 0)

        events = LazyList(lambda i: self.create_event(
            recurrence='RRULE:FREQ=DAILY;COUNT=' + str(i)), 50)

        num_submitted = 0
        num_published = 0

        for new_idx, new_event in enumerate(events):

            if new_idx > 0:

                for old_idx, old_event in enumerate(events[:new_idx - 1]):

                    if old_event.state == 'preview':

                        if (old_idx % 13 != 0):
                            old_event.submit()
                            transaction.commit()
                            num_submitted += old_idx

                    elif old_event.state == 'submitted':

                        if (old_idx % 7 != 0):
                            old_event.publish()
                            transaction.commit()
                            num_submitted -= old_idx
                            num_published += old_idx
                        else:
                            old_event.do_action("deny")
                            transaction.commit()
                            num_submitted -= old_idx

                    elif old_event.state == 'published':

                        if (old_idx % 5 != 0):
                            old_event.archive()
                            transaction.commit()
                            num_published -= old_idx

                    elif old_event.state == 'archived':

                        if (old_idx % 11 == 0):
                            old_event.publish()
                            transaction.commit()
                            num_published += old_idx

                    self.assertEqual(len(submitted.index), num_submitted)
                    self.assertEqual(len(published.index), num_published)

    def test_eventorder_metadata(self):

        class MockCatalog(object):

            def __init__(self, directory):
                self.directory = directory

        def generate_metadata(testindex):

            orderindex = EventOrderIndex(
                catalog=MockCatalog(self.directory), state='published',
                initial_index=testindex
            )

            orderindex.generate_metadata()
            return orderindex.get_metadata('dateindex')

        dateindex = generate_metadata([
            '12.01.01-12:00;'
        ])

        self.assertTrue(date(2012, 1, 1) in dateindex)
        self.assertEqual(dateindex[date(2012, 1, 1)], 0)
        self.assertEqual(len(dateindex), 1)

        dateindex = generate_metadata([
            '12.01.01-00:00;',
            '12.01.01-23:59;'
        ])

        self.assertTrue(date(2012, 1, 1) in dateindex)
        self.assertEqual(dateindex[date(2012, 1, 1)], 0)
        self.assertEqual(len(dateindex), 1)

        dateindex = generate_metadata([
            '12.01.01-00:00;',
            '12.01.01-00:00;',
            '12.01.02-00:00;'
        ])

        self.assertTrue(date(2012, 1, 1) in dateindex)
        self.assertTrue(date(2012, 1, 2) in dateindex)
        self.assertEqual(dateindex[date(2012, 1, 1)], 0)
        self.assertEqual(dateindex[date(2012, 1, 2)], 2)
        self.assertEqual(len(dateindex), 2)

        dateindex = generate_metadata([
            '12.01.01-00:00',
            '12.09.07-00:00',
            '12.12.31-00:00'
        ])

        self.assertEqual(dateindex[date(2012, 1, 1)], 0)
        self.assertEqual(dateindex[date(2012, 1, 2)], 1)
        self.assertEqual(dateindex[date(2012, 9, 6)], 1)
        self.assertEqual(dateindex[date(2012, 9, 7)], 1)
        self.assertEqual(dateindex[date(2012, 9, 8)], 2)
        self.assertEqual(dateindex[date(2012, 12, 30)], 2)
        self.assertEqual(dateindex[date(2012, 12, 31)], 2)
        self.assertEqual(len(dateindex), 366)

        dateindex = generate_metadata([
            '13.07.04-14:00;',
            '13.07.05-18:45;',
            '13.07.05-19:00;',
            '13.07.05-19:00;',
            '13.07.05-19:00;',
            '13.07.06-19:00;',
        ])

        self.assertTrue(date(2013, 07, 04) in dateindex)
        self.assertTrue(date(2013, 07, 05) in dateindex)
        self.assertTrue(date(2013, 07, 06) in dateindex)

        self.assertEqual(dateindex[date(2013, 07, 04)], 0)
        self.assertEqual(dateindex[date(2013, 07, 05)], 1)
        self.assertEqual(dateindex[date(2013, 07, 06)], 5)

    def test_attach_reindex_idempotence(self):

        # attaching the reindex to the request multiple times should
        # result in only one call to the reindex function on the catalog

        # because we cannot let the transaction go through in the test
        # we need to inspect the content of the transaction
        def number_of_data_managers():
            resources = transaction.get()._resources
            return len(
                [m for m in resources if isinstance(m, ReindexDataManager)]
            )

        attach_reindex_to_transaction(self.directory)
        self.assertEqual(1, number_of_data_managers())
        attach_reindex_to_transaction(self.directory)
        self.assertEqual(1, number_of_data_managers())
        attach_reindex_to_transaction(self.directory)
        self.assertEqual(1, number_of_data_managers())

    def test_by_range_timezone_regression(self):
        self.login_testuser()

        published = self.catalog.indices['published']

        self.assertEqual(len(published.index), 0)

        event = self.create_event(
            start=datetime(2014, 12, 31, 22),
            end=datetime(2014, 12, 31, 23),
            timezone='Europe/Vienna'
        )
        event.submit()
        event.publish()
        transaction.commit()

        self.assertEqual(len(published.index), 1)

        dtrange = (datetime(2015, 1, 1), datetime(2015, 12, 31))
        dtrange = [dates.as_timezone(dt, 'Europe/Vienna') for dt in dtrange]
        self.assertEqual(len(published.by_range(*dtrange)), 0)

        dtrange = (datetime(2014, 1, 1), datetime(2014, 12, 31, 23, 59))
        dtrange = [dates.as_timezone(dt, 'Europe/Vienna') for dt in dtrange]
        self.assertEqual(len(published.by_range(*dtrange)), 1)



import logging
log = logging.getLogger('seantis.dir.events')

from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite
from zope.annotation.interfaces import IAnnotations

from plone.app.theming.utils import (
    applyTheme,
    getOrCreatePersistentResourceDirectory
)

from seantis.dir.base.interfaces import IDirectoryCatalog
from seantis.dir.base.upgrades import (
    add_behavior_to_item,
    reset_images_and_attachments
)

from seantis.dir.events.submission import EventSubmissionData
from seantis.dir.events.interfaces import (
    IEventsDirectory,
    IEventsDirectoryItem
)


def setup_indexing(context):
    setup = getToolByName(context, 'portal_setup')

    profile = 'profile-seantis.dir.events:default'
    setup.runImportStepFromProfile(profile, 'typeinfo')
    setup.runImportStepFromProfile(profile, 'catalog')

    # reindex everything
    catalog = getToolByName(context, 'portal_catalog')
    catalog.clearFindAndRebuild()

    # rebuild eventindexes
    directories = catalog(object_provides=IEventsDirectory.__identifier__)

    for directory in directories:

        directory_catalog = IDirectoryCatalog(directory.getObject())
        directory_catalog.reindex()


def upgrade_1000_to_1001(context):
    add_behavior_to_item(context, 'seantis.dir.events', IEventsDirectoryItem)
    reset_images_and_attachments(
        context,
        (IEventsDirectory, IEventsDirectoryItem),
        ['image', 'attachment_1', 'attachment_2']
    )


def upgrade_1001_to_1002(context):

    # 1002 untangles the dependency hell that was default <- sunburst <- izug.
    # Now, sunburst and izug.basetheme both have their own profiles.

    # Since the default profile therefore has only the bare essential styles
    # it needs to be decided on upgrade which theme was used, the old css
    # files need to be removed and the theme profile needs to be applied.

    # acquire the current theme
    skins = getToolByName(context, 'portal_skins')
    theme = skins.getDefaultSkin()

    # find the right profile to use
    profilemap = {
        'iZug Base Theme': 'izug_basetheme',
        'Sunburst Theme': 'sunburst'
    }

    if theme not in profilemap:
        log.info("Theme %s is not supported by seantis.dir.events" % theme)
        profile = 'default'
    else:
        profile = profilemap[theme]

    # remove all existing event stylesheets
    css_registry = getToolByName(context, 'portal_css')
    stylesheets = css_registry.getResourcesDict()
    ids = [i for i in stylesheets if 'resource++seantis.dir.events.css' in i]

    map(css_registry.unregisterResource, ids)

    # remove the old diazo theme and disable theming for now
    themes = getOrCreatePersistentResourceDirectory()
    to_delete = [
        'izug.seantis.dir.events.theme',
        'vbeo.seantis.dir.events.theme'
    ]
    for theme in to_delete:
        try:
            del themes[theme]
        except KeyError:
            continue

    applyTheme(None)

    setup = getToolByName(context, 'portal_setup')

    # reapply the chosen profile
    setup.runAllImportStepsFromProfile(
        'profile-seantis.dir.events:%s' % profile
    )

    # there are currently two installations, bern and zug
    if getSite().id == 'vbeo':
        setup.runAllImportStepsFromProfile(
            'profile-vbeo.seantis.dir.events:default'
        )
    else:
        setup.runAllImportStepsFromProfile(
            'profile-izug.seantis.dir.events:default'
        )


def upgrade_1002_to_1003(context):
    # the new plone.formwidget.recurrence release introduced gs profiles which
    # need to be loaded on existing sites

    setup = getToolByName(context, 'portal_setup')
    setup.runAllImportStepsFromProfile(
        'profile-plone.formwidget.recurrence:default'
    )


def upgrade_1003_to_1004(context):
    # import new javascript
    setup = getToolByName(context, 'portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.dir.events:default', 'jsregistry'
    )
    setup.runImportStepFromProfile(
        'profile-seantis.dir.events:default', 'cssregistry'
    )


def upgrade_1004_to_1005(context):
    # adds a new behvaior
    setup = getToolByName(context, 'portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.dir.events:default', 'typeinfo'
    )

    # add underscore.js
    setup.runAllImportStepsFromProfile(
        'profile-collective.js.underscore:default'
    )

    # upgrade javascript
    setup.runImportStepFromProfile(
        'profile-seantis.dir.events:default', 'jsregistry'
    )

    # guess the submission data for the existing events. This does
    # not actually lead to a change of the event dates, it just opens
    # the possiblity to do so by manually editing and saving the event
    catalog = getToolByName(context, 'portal_catalog')
    brains = catalog(object_provides=IEventsDirectoryItem.__identifier__)

    h = 60 * 60

    for brain in brains:
        # don't acquire through interface as that would require more reindexing
        event = brain.getObject()
        submission = EventSubmissionData(event)

        if (event.end - event.start).total_seconds() < 24 * h:
            submission.submission_date_type = ['date']
            submission.submission_date = event.local_start.date()
            submission.submission_start_time = event.local_start.time()
            submission.submission_end_time = event.local_end.time()
            submission.submission_recurrence = event.recurrence
        else:
            # eventual recurrence is lost here, it's no longer possible
            # to have recurring events which last for days
            submission.submission_date_type = ['range']
            submission.submission_range_start_date = event.local_start.date()
            submission.submission_range_end_date = event.local_end.date()
            submission.submission_range_start_time = event.local_start.time()
            submission.submission_range_end_time = event.local_end.time()

        # the guidle import had bug where whole_day was not set at all times
        if hasattr(event, 'whole_day'):
            submission.submission_whole_day = event.whole_day
        else:
            submission.submission_whole_day = False


def upgrade_1005_to_1006(context):
    setup = getToolByName(context, 'portal_setup')

    profiles = [
        'teamraum', 'sunburst', 'izug_basetheme'
    ]

    for profile in profiles:
        full_profile = 'profile-seantis.dir.events:{}'.format(profile)

        if setup.getProfileImportDate(full_profile):
            setup.runImportStepFromProfile(full_profile, 'cssregistry')


def upgrade_1006_to_1007(context):
    setup = getToolByName(context, 'portal_setup')

    # For some reason this profile was not imported on some older events
    # installations. This leads to exceptions.
    setup.runImportStepFromProfile(
        'profile-plone.app.event:default', 'browserlayer'
    )


def upgrade_1007_to_1008(context):
    # add collective.geo.behaviour
    setup = getToolByName(context, 'portal_setup')
    setup.runAllImportStepsFromProfile(
        'profile-collective.geo.behaviour:default'
    )

    add_behavior_to_item(
        context, 'seantis.dir.events', IEventsDirectoryItem
    )

    # update css and js
    getToolByName(context, 'portal_css').cookResources()
    getToolByName(context, 'portal_javascripts').cookResources()


def upgrade_1008_to_1009(context):
    # update css
    getToolByName(context, 'portal_css').cookResources()


def upgrade_1009_to_1010(context):
    # update css
    getToolByName(context, 'portal_css').cookResources()

    # add new js files
    setup = getToolByName(context, 'portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.dir.events:default', 'jsregistry'
    )


def upgrade_1010_to_1011(context):
    # Add new types and workflow
    setup = getToolByName(context, 'portal_setup')
    profile = 'profile-seantis.dir.events:default'
    setup.runImportStepFromProfile(profile, 'typeinfo')
    setup.runImportStepFromProfile(profile, 'workflow')

    # Add source to index
    setup.runImportStepFromProfile(profile, 'catalog')
    catalog = getToolByName(context, 'portal_catalog')
    if 'source' not in catalog.indexes():
        catalog.addIndex('source', 'FieldIndex')
        catalog.manage_reindexIndex(ids=['source'])


def upgrade_1011_to_1012(context):
    # import pdb; pdb.set_trace()

    # Remove old annotations from IEventsDirectory
    catalog = getToolByName(context, 'portal_catalog')
    brains = catalog(object_provides=IEventsDirectory.__identifier__)

    names = (
        'eventorder-hidden1.0',
        'eventorder-hidden1.0_meta_dateindex',
        'eventorder-published1.0',
        'eventorder-published1.0_meta_dateindex',
        'eventorder-submitted1.0',
        'eventorder-submitted1.0_meta_dateindex'
    )

    for brain in brains:
        annotations = IAnnotations(brain.getObject())
        for name in names:
            if name in annotations:
                try:
                    del annotations[name]
                except:
                    pass



import mock
import pytz

from datetime import datetime, timedelta

from collective.geo.geographer.interfaces import IGeoreferenced
from plone.app.event.base import default_timezone
from plone.dexterity.utils import createContentInContainer
from zope.interface import alsoProvides

from seantis.dir.events.catalog import reindex_directory
from seantis.dir.events.dates import default_now
from seantis.dir.events.sources import (
    import_scheduler, ExternalEventImporter, IExternalEvent
)
from seantis.dir.events.sources.guidle import EventsSourceGuidle
from seantis.dir.events.sources.seantis_json import EventsSourceSeantisJson
from seantis.dir.events.tests import IntegrationTestCase

import transaction


class TestImport(IntegrationTestCase):

    def setUp(self):
        super(TestImport, self).setUp()
        # login to gain the right to create events
        self.login_admin()

    def tearDown(self):
        super(TestImport, self).tearDown()
        self.logout()

    def create_guidle_source(self, **kw):
        """ Create a guididle source item in self.directory.
        By default, test-user must
        be logged in or the creation will be unauthorized.

        """

        defaults = {
            'source': '',
            'enabled': True
        }

        for attr in defaults:
            if not attr in kw:
                kw[attr] = defaults[attr]

        return createContentInContainer(
            self.directory, 'seantis.dir.events.sourceguidle', **kw
        )

    def cleanup_after_fetch_one(self):
        """ fetch_one does a transaction.commit which causes the test folder
        and imported events to be persistent.
        """
        self.directory.manage_delObjects(self.directory.keys())
        self.portal.manage_delObjects([self.directory.id])
        transaction.commit()

    def create_fetch_entry(self, **kw):
        defaults = {
            # from IDirectoryItem
            # description not used
            'title': '',

            # from IEventsDirectoryItem
            # submitter, submitter_email not used
            'short_description': '',
            'long_description': '',
            'image': '',
            'attachment_1': '',
            'attachment_2': '',
            'locality': '',
            'street': '',
            'housenumber': '',
            'zipcode': '',
            'town': '',
            'location_url': '',
            'event_url': '',
            'organizer': '',
            'contact_name': '',
            'contact_email': '',
            'contact_phone': '',
            'prices': '',
            'registration': '',

            # from IExternalEvent
            # source not used (filled in by ExternalEventImporter)
            'source_id': 'id-1',

            # From IEventBasic
            'start': datetime.today() + timedelta(days=10),
            'end': datetime.today() + timedelta(days=10, hours=1),
            'whole_day': False,
            'timezone': default_timezone(),

            # From IEventRecurrence
            'recurrence': '',

            # additional attributes used to control import
            'fetch_id': 'fetch-1',
            'last_update': default_now().replace(microsecond=0),
            # 'latitude': '',
            # 'longitude': '',
            'cat1': set(),
            'cat2': set(),
        }

        for attr in defaults:
            if not attr in kw:
                kw[attr] = defaults[attr]

        return kw

    def test_importer_sources(self):
        self.create_guidle_source(enabled=True)
        self.create_guidle_source(enabled=False)

        self.assertEqual(
            len(ExternalEventImporter(self.directory).sources()), 1)

    def test_importer_switch_indexing(self):
        importer = ExternalEventImporter(self.directory)

        event = self.create_event()
        event.submit()
        event.publish()
        reindex_directory(self.directory)
        self.assertEquals(len(self.catalog.ix_published.index), 1)

        importer.disable_indexing()
        event = self.create_event()
        event.submit()
        event.publish()
        reindex_directory(self.directory)
        self.assertEquals(len(self.catalog.ix_published.index), 1)

        importer.enable_indexing()
        reindex_directory(self.directory)
        self.assertEquals(len(self.catalog.ix_published.index), 2)

    def test_importer_update_time(self):
        importer = ExternalEventImporter(self.directory)

        set_time = lambda t: importer.set_last_update_time(t)
        get_time = lambda: importer.get_last_update_time()

        # No key
        self.assertRaises(AttributeError, get_time)
        self.assertRaises(AttributeError, set_time, default_now())
        self.assertRaises(AttributeError, get_time)

        importer.annotation_key = 'key'

        # Wrong dates
        self.assertRaises(AssertionError, set_time, None)
        self.assertRaises(AssertionError, set_time, 25)
        self.assertRaises(AssertionError, set_time, datetime.today())

        # Ok
        update = default_now()
        importer.set_last_update_time(update)
        last_update = importer.get_last_update_time()
        self.assertEquals(last_update, update.replace(microsecond=0))

    def test_importer_existing_events(self):
        importer = ExternalEventImporter(self.directory)

        sources = 2 * ['source1', 'source1', 'source2', '', None, 'source-3']
        for source in sources:
            args = {} if source is None else {'source': source}
            event = self.create_event(**args)
            event.submit()
            event.publish()
            if source is not None:
                alsoProvides(event, IExternalEvent)
            event.reindexObject()

        for source in list(set(sources)):
            if source is not None:
                self.assertEquals(len(importer.existing_events(source)),
                                  sources.count(source))

    def test_importer_group_by_source_id(self):
        importer = ExternalEventImporter(self.directory)

        ids = ['1', '101', '27', '100', '2', '100', '27', '1']
        events = [self.create_event(source_id=id) for id in ids]

        groups = importer.groupby_source_id(events)
        for id in groups:
            self.assertEquals(len(groups[id]), ids.count(id))

    def test_importer_fetch_one(self):
        try:
            importer = ExternalEventImporter(self.directory)
            events = []
            fetch = lambda: events
            from_ids = lambda ids: [self.create_fetch_entry(source_id=id,
                                                            fetch_id='f')
                                    for id in ids]

            # Simple import
            ids = ['event1', 'event2', 'event3', 'event4',
                   'event5', 'event6', 'event7', 'event8']
            events = from_ids(ids[:4])
            imports = importer.fetch_one('source', fetch)
            self.assertEquals(imports, 4)
            imported = [i.getObject().source_id for i in self.catalog.query()]
            self.assertEquals(ids[:4], imported)

            # Import with limit
            events = from_ids(ids[4:])
            imports = importer.fetch_one('source', fetch, limit=2)
            self.assertEquals(imports, 2)
            self.assertEquals(len(self.catalog.query()), 6)
            imports = importer.fetch_one('source', fetch)
            self.assertEquals(imports, 2)
            self.assertEquals(len(self.catalog.query()), 8)

            # Force reimport
            imports = importer.fetch_one('source', fetch)
            self.assertEquals(imports, 0)
            self.assertEquals(len(self.catalog.query()), 8)
            imports = importer.fetch_one('source', fetch,
                                                  reimport=True)
            self.assertEquals(imports, 4)
            self.assertEquals(len(self.catalog.query()), 8)

            # Reimport updated events
            events = from_ids(ids)
            imports = importer.fetch_one('source', fetch)
            self.assertEquals(imports, 8)
            self.assertEquals(len(self.catalog.query()), 8)

            # Test import of given source IDs only
            events = from_ids(ids)
            imports = importer.fetch_one(
                'source', fetch, source_ids=ids[2:6]
            )
            self.assertEquals(imports, 4)
            self.assertEquals(len(self.catalog.query()), 8)

        finally:
            # Clean up (transaction has been commited)
            self.cleanup_after_fetch_one()

    def test_importer_update_category_suggestions(self):
        try:
            importer = ExternalEventImporter(self.directory)

            events = []
            fetch = lambda: events

            events.append(self.create_fetch_entry(
                source_id='1', fetch_id='1',
                cat1=set(['cat1-1']), cat2=set(['cat2-1'])
            ))
            events.append(self.create_fetch_entry(
                source_id='2', fetch_id='1',
                cat1=set(['cat1-2', 'cat1-4']), cat2=set(['cat2-1'])
            ))
            events.append(self.create_fetch_entry(
                source_id='3', fetch_id='1',
                cat1=set(), cat2=set(['cat2-1', 'cat2-2', 'cat2-3'])
            ))

            imports = importer.fetch_one('source', fetch)
            self.assertEquals(imports, 3)
            self.assertTrue('cat1-1' in self.directory.cat1_suggestions)
            self.assertTrue('cat1-2' in self.directory.cat1_suggestions)
            self.assertTrue('cat1-4' in self.directory.cat1_suggestions)
            self.assertTrue('cat2-1' in self.directory.cat2_suggestions)
            self.assertTrue('cat2-2' in self.directory.cat2_suggestions)
            self.assertTrue('cat2-3' in self.directory.cat2_suggestions)

        finally:
            # Clean up (transaction has been commited)
            self.cleanup_after_fetch_one()

    def test_importer_values(self):
        try:
            importer = ExternalEventImporter(self.directory)

            string_values = [
                'title', 'short_description', 'long_description',
                'locality', 'street', 'housenumber', 'zipcode', 'town',
                'location_url', 'event_url', 'organizer',
                'contact_name', 'contact_email', 'contact_phone',
                'prices', 'registration',
                'source_id', 'fetch_id'
            ]
            now = default_now().replace(microsecond=0)
            then = now + timedelta(days=10)

            event = {s: s for s in string_values}

            event['last_update'] = now
            event['start'] = then
            event['end'] = then + timedelta(hours=1)
            event['timezone'] = default_timezone()
            event['whole_day'] = False
            event['recurrence'] = 'RRULE:FREQ=DAILY;COUNT=2'

            event['cat1'] = set(['c1', 'c2'])
            event['cat2'] = set(['c3', 'c4'])

            event['longitude'] = 7.8673189
            event['latitude'] = 46.6859853

            # :TODO: test image and attachement download
            # event['image'] =
            # event['attachment_1'] =
            # event['attachment_2'] =

            imports = importer.fetch_one('source', lambda: [event])
            imported = self.catalog.query()
            self.assertEquals(imports, 1)
            self.assertEquals(len(imported), 1)

            imported = imported[0].getObject()
            for s in string_values:
                self.assertTrue(s in vars(imported))
                self.assertTrue(vars(imported)[s] == s)

            self.assertEquals(imported.start, now + timedelta(days=10))
            self.assertEquals(imported.end, now + timedelta(days=10, hours=1))
            self.assertEquals(imported.recurrence, 'RRULE:FREQ=DAILY;COUNT=2')
            self.assertFalse(imported.whole_day)

            self.assertTrue('c1' in imported.cat1)
            self.assertTrue('c2' in imported.cat1)
            self.assertTrue('c3' in imported.cat2)
            self.assertTrue('c4' in imported.cat2)

            self.assertEquals(
                IGeoreferenced(imported).coordinates, [7.8673189, 46.6859853]
            )

        finally:
            # Clean up (transaction has been commited)
            self.cleanup_after_fetch_one()

    def test_importer_autoremove(self):
        try:
            importer = ExternalEventImporter(self.directory)

            events = []
            fetch = lambda: events

            # First import
            events.append(self.create_fetch_entry(source_id='source_id_1',
                                                  fetch_id='fetch_id'))
            events.append(self.create_fetch_entry(source_id='source_id_1',
                                                  fetch_id='fetch_id'))
            events.append(self.create_fetch_entry(source_id='source_id_2',
                                                  fetch_id='fetch_id'))
            imports = importer.fetch_one('source', fetch)
            self.assertEquals(imports, 3)
            self.assertEquals(len(self.directory.keys()), 3)

            # Second import
            events.pop(0)
            events.pop(0)
            events.append(self.create_fetch_entry(source_id='source_id_3',
                                                  fetch_id='fetch_id'))

            imports = importer.fetch_one('source', fetch,
                                                  autoremove=True)
            self.assertEquals(imports, 1)

            ids = [item.source_id for id, item in self.directory.objectItems()]
            self.assertEquals(len(ids), 2)
            self.assertTrue('source_id_1' not in ids)

        finally:
            # Clean up (transaction has been commited)
            self.cleanup_after_fetch_one()

    def test_importer_keep_hidden(self):
        try:
            # Import event
            importer = ExternalEventImporter(self.directory)
            event = self.create_fetch_entry(source_id='s', fetch_id='f')
            imports = importer.fetch_one('source', lambda: [event])
            self.assertEquals(imports, 1)

            # Hide event
            brains = self.catalog.catalog(
                object_provides=IExternalEvent.__identifier__
            )
            self.assertEquals(len(brains), 1)
            hidden = brains[0].getObject()
            hidden.hide()

            # Re-import event
            imports = importer.fetch_one('source', lambda: [event])
            self.assertEquals(imports, 0)
            imports = importer.fetch_one(
                'source', lambda: [event], reimport=True
            )
            self.assertEquals(imports, 1)

            brains = self.catalog.catalog(
                object_provides=IExternalEvent.__identifier__
            )
            self.assertEquals(len(brains), 1)
            event = brains[0].getObject()
            self.assertTrue(
                event.modification_date != hidden.modification_date
            )
            self.assertEquals(event.review_state, 'hidden')

        finally:
            # Clean up (transaction has been commited)
            self.cleanup_after_fetch_one()

    def test_importer_export_imported(self):
        try:
            # Import event
            importer = ExternalEventImporter(self.directory)
            event = self.create_fetch_entry(source_id='s', fetch_id='f')
            imports = importer.fetch_one('source', lambda: [event])
            self.assertEquals(imports, 1)

            # Add own event
            event = self.create_event(start=datetime.today().replace(second=0))
            event.submit()
            event.publish()
            reindex_directory(self.directory)

            # Export events
            events = [idx for idx in enumerate(self.catalog.export())]
            self.assertEquals(len(events), 1)

        finally:
            # Clean up (transaction has been commited)
            self.cleanup_after_fetch_one()

    def test_guidle_import(self):
        xml = """<?xml version="1.0" encoding="UTF-8"?>
<guidle:exportData xmlns:guidle="http://www.guidle.com">
  <guidle:groupSet>
    <guidle:group>
      <guidle:offer id="123">
        <guidle:lastUpdateDate>2013-07-26T16:00:43.208+02:00</guidle:lastUpdateDate>
        <guidle:offerDetail>
          <guidle:title>Title</guidle:title>
          <guidle:shortDescription>Description</guidle:shortDescription>
          <guidle:longDescription>Long Descritption</guidle:longDescription>
          <guidle:homepage>http://www.example.ch</guidle:homepage>
          <guidle:images>
            <guidle:image>
              <guidle:url>http://www.example.ch/1.png</guidle:url>
            </guidle:image>
          </guidle:images>
        </guidle:offerDetail>
        <guidle:address>
          <guidle:company>Address</guidle:company>
          <guidle:zip>1234</guidle:zip>
          <guidle:city>City</guidle:city>
          <guidle:country>Country</guidle:country>
          <guidle:latitude>1.0</guidle:latitude>
          <guidle:longitude>2.0</guidle:longitude>
        </guidle:address>
        <guidle:contact>
          <guidle:email>info@example.ch</guidle:email>
          <guidle:telephone_1>000 111 22 33</guidle:telephone_1>
          <guidle:company>Company</guidle:company>
        </guidle:contact>
        <guidle:schedules>
          <guidle:date>
            <guidle:startDate>2017-08-25</guidle:startDate>
            <guidle:endDate>2017-09-03</guidle:endDate>
            <guidle:weekdays>
              <guidle:day>Mo</guidle:day>
              <guidle:day>Tu</guidle:day>
            </guidle:weekdays>
          </guidle:date>
        </guidle:schedules>
        <guidle:classifications>
          <guidle:classification name="class">
          </guidle:classification>
        </guidle:classifications>
      </guidle:offer>
      <guidle:offer id="234">
        <guidle:lastUpdateDate>2013-07-26T16:00:43.208+02:00</guidle:lastUpdateDate>
        <guidle:offerDetail>
          <guidle:title>Title</guidle:title>
          <guidle:shortDescription>Description</guidle:shortDescription>
          <guidle:longDescription>Long Descritption</guidle:longDescription>
          <guidle:homepage>http://www.example.ch</guidle:homepage>
          <guidle:images>
            <guidle:image>
              <guidle:url>http://www.example.ch/1.png</guidle:url>
            </guidle:image>
          </guidle:images>
        </guidle:offerDetail>
        <guidle:address>
          <guidle:company>Address</guidle:company>
          <guidle:zip>1234</guidle:zip>
          <guidle:city>City</guidle:city>
          <guidle:country>Country</guidle:country>
        </guidle:address>
        <guidle:contact>
          <guidle:email>info@example.ch</guidle:email>
          <guidle:telephone_1>000 111 22 33</guidle:telephone_1>
          <guidle:company>Company</guidle:company>
        </guidle:contact>
        <guidle:schedules>
          <guidle:date>
            <guidle:startDate>2017-08-25</guidle:startDate>
            <guidle:endDate>2017-08-25</guidle:endDate>
            <guidle:startTime>09:15:00</guidle:startTime>
            <guidle:endTime>20:50:00</guidle:endTime>
          </guidle:date>
        </guidle:schedules>
        <guidle:classifications>
          <guidle:classification name="class">
          </guidle:classification>
        </guidle:classifications>
      </guidle:offer>
    </guidle:group>
  </guidle:groupSet>
</guidle:exportData>"""

        context = mock.Mock()
        context.url = 'url'

        source = EventsSourceGuidle(context)
        events = [event for event in source.fetch(xml)]
        self.assertEquals(len(events), 2)

        self.assertEquals(str(events[0]['last_update']),
                          '2013-07-26 16:00:43+02:00')
        self.assertEquals(events[0]['fetch_id'], 'url')
        self.assertEquals(events[0]['id'], '123')
        self.assertEquals(events[0]['source_id'], '123')
        self.assertEquals(events[0]['title'], 'Title')
        self.assertEquals(events[0]['short_description'], 'Description')
        self.assertEquals(events[0]['long_description'], 'Long Descritption')
        self.assertEquals(events[0]['location_url'], 'http://www.example.ch')
        self.assertEquals(events[0]['image'], 'http://www.example.ch/1.png')
        self.assertEquals(events[0]['organizer'], 'Company')
        self.assertEquals(events[0]['locality'], 'Address')
        self.assertEquals(events[0]['zipcode'], '1234')
        self.assertEquals(events[0]['town'], 'City')
        self.assertEquals(events[0]['contact_email'], 'info@example.ch')
        self.assertEquals(events[0]['contact_phone'], '000 111 22 33')
        self.assertEquals(events[0]['latitude'], '1.0')
        self.assertEquals(events[0]['longitude'], '2.0')
        self.assertEquals(events[0]['start'], datetime(2017, 8, 25, 0, 0))
        self.assertEquals(events[0]['end'], datetime(2017, 8, 25, 0, 0))
        self.assertEquals(events[0]['recurrence'],
                          'RRULE:FREQ=WEEKLY;BYDAY=MO,TU;UNTIL=20170904T0000Z')
        self.assertEquals(events[0]['whole_day'], True)
        self.assertEquals(events[0]['timezone'], 'Europe/Zurich')
        self.assertEquals(events[0]['cat1'], set())
        self.assertEquals(events[0]['cat2'], set(['City']))

        self.assertEquals(str(events[1]['last_update']),
                          '2013-07-26 16:00:43+02:00')
        self.assertEquals(events[1]['fetch_id'], 'url')
        self.assertEquals(events[1]['id'], '234')
        self.assertEquals(events[1]['source_id'], '234')
        self.assertEquals(events[1]['title'], 'Title')
        self.assertEquals(events[1]['short_description'], 'Description')
        self.assertEquals(events[1]['long_description'], 'Long Descritption')
        self.assertEquals(events[1]['location_url'], 'http://www.example.ch')
        self.assertEquals(events[1]['image'], 'http://www.example.ch/1.png')
        self.assertEquals(events[1]['organizer'], 'Company')
        self.assertEquals(events[1]['locality'], 'Address')
        self.assertEquals(events[1]['zipcode'], '1234')
        self.assertEquals(events[1]['town'], 'City')
        self.assertEquals(events[1]['contact_email'], 'info@example.ch')
        self.assertEquals(events[1]['contact_phone'], '000 111 22 33')
        self.assertTrue('latitude' not in events[1])
        self.assertTrue('longitude' not in events[1])
        self.assertEquals(events[1]['start'], datetime(2017, 8, 25, 9, 15))
        self.assertEquals(events[1]['end'], datetime(2017, 8, 25, 20, 50))
        self.assertEquals(events[1]['recurrence'], '')
        self.assertEquals(events[1]['whole_day'], False)
        self.assertEquals(events[1]['timezone'], 'Europe/Zurich')
        self.assertEquals(events[1]['cat1'], set())
        self.assertEquals(events[1]['cat2'], set(['City']))

    def test_seantis_import(self):
        json_string = """[{
            "id": "id1", "title": "title",
            "short_description": "short_description",
            "long_description": "h\u00e4nsel",
            "cat1": ["cat12", "cat11"], "cat2": ["cat21"],
            "event_url": "http://www.event.ch",
            "timezone": "Europe/Zurich",
            "start": "2014-01-15T00:00:00+00:00",
            "end": "2014-01-15T23:59:59+00:00",
            "whole_day": true,
            "recurrence": "RRULE:FREQ=WEEKLY;BYDAY=MO,TU;UNTIL=20150101T0000Z",
            "locality": "locality",
            "street": "street", "housenumber": "housenumber",
            "zipcode": "1234", "town": "town",
            "longitude": 2.0, "latitude": 1.0,
            "location_url": "http://www.location.ch",
            "contact_name": "contact_name",
            "contact_phone": "+12 (3) 45 678 90 12",
            "contact_email": "contact@ema.il",
            "registration": "http://www.tickets.ch",
            "organizer": "organizer",
            "prices": "prices",
            "images": [{"url": "img_url", "name": "img_name"}],
            "attachements": [
                {"url": "a1_url", "name": "a1_name"},
                {"url": "a2_url", "name": "a2_name"}
            ],
            "submitter": "sumitter", "submitter_email": "submitter@ma.il"
        },{
            "last_update": "2014-01-21T10:21:47+01:00",
            "id": "test", "title": "test",
            "short_description": "test", "long_description": null,
            "cat1": ["cat13", "cat14"], "cat2": ["cat21", "cat22", "cat23"],
            "event_url": null,
            "timezone": "UTC",
            "start": "2014-01-19T17:00:00+02:00",
            "end": "2014-01-19T18:00:00+02:00",
            "whole_day": false, "recurrence": null,
            "locality": null, "street": null, "housenumber": null,
            "zipcode": null, "town": null,
            "longitude": null, "latitude": null,
            "location_url": null,
            "contact_name": null, "contact_phone": null, "contact_email": null,
            "registration": null, "organizer": null, "prices": null,
            "images": null, "attachments": [],
            "submitter": "cccc", "submitter_email": "submitter@ma.il"
        }]"""

        context = mock.Mock()
        context.url = 'url'
        context.do_filter = False
        context.cat1 = ''
        context.cat2 = ''

        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 2)

        self.assertTrue(
            default_now() - events[0]['last_update'] < timedelta(seconds=10)
        )
        self.assertEquals(events[0]['fetch_id'], 'url')
        self.assertEquals(events[0]['id'], 'id1')
        self.assertEquals(events[0]['source_id'], 'id1')
        self.assertEquals(events[0]['title'], 'title')
        self.assertEquals(events[0]['short_description'], u'short_description')
        self.assertEquals(events[0]['long_description'], u'h\xe4nsel')
        self.assertEquals(events[0]['event_url'], 'http://www.event.ch')
        self.assertEquals(events[0]['registration'], 'http://www.tickets.ch')
        self.assertEquals(events[0]['location_url'], 'http://www.location.ch')
        self.assertEquals(events[0]['image'], 'img_url')
        self.assertEquals(events[0]['image_name'], 'img_name')
        self.assertEquals(events[0]['attachment_1'], 'a1_url')
        self.assertEquals(events[0]['attachment_1_name'], 'a1_name')
        self.assertEquals(events[0]['attachment_2'], 'a2_url')
        self.assertEquals(events[0]['attachment_2_name'], 'a2_name')
        self.assertEquals(events[0]['organizer'], 'organizer')
        self.assertEquals(events[0]['street'], 'street')
        self.assertEquals(events[0]['housenumber'], 'housenumber')
        self.assertEquals(events[0]['locality'], 'locality')
        self.assertEquals(events[0]['zipcode'], '1234')
        self.assertEquals(events[0]['town'], 'town')
        self.assertEquals(events[0]['contact_name'], 'contact_name')
        self.assertEquals(events[0]['contact_email'], 'contact@ema.il')
        self.assertEquals(events[0]['contact_phone'], '+12 (3) 45 678 90 12')
        self.assertEquals(events[0]['latitude'], '1.0')
        self.assertEquals(events[0]['longitude'], '2.0')
        self.assertEquals(events[0]['timezone'], 'Europe/Zurich')
        self.assertEquals(events[0]['start'], datetime(2014, 1, 15, 0, 0,
                                                       tzinfo=pytz.UTC))
        self.assertEquals(events[0]['end'], datetime(2014, 1, 15, 23, 59, 59,
                                                     tzinfo=pytz.UTC))
        self.assertEquals(events[0]['recurrence'],
                          'RRULE:FREQ=WEEKLY;BYDAY=MO,TU;UNTIL=20150101T0000Z')
        self.assertEquals(events[0]['whole_day'], True)
        self.assertEquals(events[0]['cat1'], set(['cat11', 'cat12']))
        self.assertEquals(events[0]['cat2'], set(['cat21']))
        self.assertEquals(events[0]['submitter'], 'sumitter')
        self.assertEquals(events[0]['submitter_email'], 'submitter@ma.il')

        self.assertEquals(
            events[1]['last_update'],
            datetime(2014, 1, 21, 9, 21, 47, tzinfo=pytz.UTC)
        )
        self.assertEquals(events[1]['latitude'], None)
        self.assertEquals(events[1]['longitude'], None)
        self.assertEquals(events[1]['timezone'], 'UTC')
        self.assertEquals(events[1]['start'], datetime(2014, 1, 19, 15, 0,
                                                       tzinfo=pytz.UTC))
        self.assertEquals(events[1]['end'], datetime(2014, 1, 19, 16, 0,
                                                     tzinfo=pytz.UTC))
        self.assertEquals(events[1]['whole_day'], False)
        self.assertEquals(events[1]['cat1'], set(['cat13', 'cat14']))
        self.assertEquals(events[1]['cat2'], set(['cat21', 'cat22', 'cat23']))

        # Filter by categories
        context.do_filter = False
        context.cat1 = 'cat5'
        context.cat2 = 'cat6'
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 2)

        context.do_filter = True

        context.cat1 = ''
        context.cat2 = ''
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 2)

        context.cat1 = 'cat1'
        context.cat2 = ''
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 0)

        context.cat1 = 'cat11'
        context.cat2 = ''
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['cat1'], set(['cat11', 'cat12']))

        context.cat1 = 'cat12'
        context.cat2 = ''
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['cat1'], set(['cat11', 'cat12']))

        context.cat1 = ''
        context.cat2 = 'cat23'
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['cat2'], set(['cat21', 'cat22', 'cat23']))

        context.cat1 = ''
        context.cat2 = 'cat24'
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 0)

        context.do_filter = True
        context.cat1 = 'cat11'
        context.cat2 = 'cat21'
        source = EventsSourceSeantisJson(context)
        events = [event for event in source.fetch(json_string)]
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['cat1'], set(['cat11', 'cat12']))
        self.assertEquals(events[0]['cat2'], set(['cat21']))

    def test_seantis_import_build_url(self):
        context = mock.Mock()
        context.url = '  http://www.ex.ch/ev/  '
        context.do_filter = False
        context.cat1 = ''
        context.cat2 = ''
        source = EventsSourceSeantisJson(context)
        self.assertEquals(
            source.build_url(),
            'http://www.ex.ch/ev/?type=json&compact=true'
        )

        context.url = '  http://www.ex.ch/ev  '
        source = EventsSourceSeantisJson(context)
        self.assertEquals(
            source.build_url(),
            'http://www.ex.ch/ev?type=json&compact=true'
        )

        context.do_filter = True
        source = EventsSourceSeantisJson(context)
        self.assertEquals(
            source.build_url(),
            'http://www.ex.ch/ev?type=json&compact=true'
        )

        context.cat1 = 'cat1'
        source = EventsSourceSeantisJson(context)
        self.assertEquals(
            source.build_url(),
            'http://www.ex.ch/ev?type=json&compact=true&filter=true&cat1=cat1'
        )

        context.cat1 = ''
        context.cat2 = 'cat2'
        source = EventsSourceSeantisJson(context)
        self.assertEquals(
            source.build_url(),
            'http://www.ex.ch/ev?type=json&compact=true&filter=true&cat2=cat2'
        )

        context.cat1 = 'cat1'
        source = EventsSourceSeantisJson(context)
        self.assertEquals(
            source.build_url(),
            'http://www.ex.ch/ev?type=json&compact=true&filter=true'
            '&cat1=cat1&cat2=cat2'
        )



import logging
log = logging.getLogger('seantis.dir.events')

import imghdr
import magic

from collective.dexteritytextindexer import searchable
from plone.namedfile.field import NamedImage, NamedFile
from plone.app.event.dx.behaviors import first_weekday_sun0
from plone.directives import form
from plone.app.z3cform.wysiwyg import WysiwygFieldWidget
from plone.autoform.interfaces import IFormFieldProvider
from plone.formwidget.recurrence.z3cform.widget import RecurrenceWidget
from zope.schema import Text, TextLine, Bool, List, Choice, Time, Date, Int
from zope.interface import Invalid, Interface, Attribute, alsoProvides
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from seantis.plonetools.schemafields import Email, Website

from seantis.dir.base.interfaces import (
    IDirectory,
    IDirectoryItem,
    IDirectoryItemCategories
)

from seantis.dir.events import _


days = SimpleVocabulary(
    [
        SimpleTerm(value='MO', title=_(u'Mo')),
        SimpleTerm(value='TU', title=_(u'Tu')),
        SimpleTerm(value='WE', title=_(u'We')),
        SimpleTerm(value='TH', title=_(u'Th')),
        SimpleTerm(value='FR', title=_(u'Fr')),
        SimpleTerm(value='SA', title=_(u'Sa')),
        SimpleTerm(value='SU', title=_(u'Su')),
    ]
)

# cannot use the comprehension in the zope interface definition
# because zope takes some weird magic to work
days.keys = [str(d.value) for d in days._terms]


class ITokenAccess(Interface):

    def attach_token(self, token=None):
        "Optionally create and store the access token."

    def has_access(self, request):
        "Return true if the given request has access on the context"

    def clear_token(self):
        "Remove the token from context and session."


class IActionGuard(Interface):

    def allow_action(self, action, item):
        """Return true if the given workflow_events action is allowed.
        The item can currently be a brain or not which is fugly, so maybe
        don't use for now and wait for a better implementation."""


class IExternalEvent(Interface):

    source = Attribute("Name of external source -> seantis.dir.events.sources")
    source_id = Attribute("""
        Id of external source event. External sources may use the same id
        for multiple events. If the id already exists when fetching the source
        the source is used and the existing events with the same id are deleted
    """)


class IEventsDirectory(IDirectory):
    """Extends the seantis.dir.base.directory.IDirectory"""

    image = NamedImage(
        title=_(u'Image'),
        required=False,
        default=None
    )

    searchable('terms')
    terms = Text(
        title=_(u'Terms and Conditions'),
        description=_(
            u'If entered, the terms and conditions have '
            u'to be agreed to by anyone submitting an event.'
        ),
        required=False,
        default=None
    )
    form.widget(terms=WysiwygFieldWidget)


# Hide all categories as they are predefined
IEventsDirectory.setTaggedValue(
    'seantis.dir.base.omitted',
    [
        'cat1',
        'cat2',
        'cat3',
        'cat4',
        'cat3_suggestions',
        'cat4_suggestions',
        'cat1_descriptions',
        'cat2_descriptions',
        'cat3_descriptions',
        'cat4_descriptions',
        'allow_custom_categories'  # always locked
    ]
)

IEventsDirectory.setTaggedValue('seantis.dir.base.labels', {
    'cat1_suggestions': _("Suggested Values for the What-Category"),
    'cat2_suggestions': _("Suggested Values for the Where-Category"),
})


class IEventsDirectoryItem(IDirectoryItem):
    """Extends the seantis.dir.IDirectoryItem."""

    submitter = TextLine(
        title=_(u'Submitter Name'),
        required=False,
    )

    submitter_email = Email(
        title=_(u'Submitter Email'),
        required=False,
    )

    searchable('short_description')
    short_description = Text(
        title=_(u'Short Description'),
        description=_(u'Up to 140 characters'),
        required=True,
        max_length=140
    )

    searchable('long_description')
    long_description = Text(
        title=_(u'Long Description'),
        required=False
    )

    image = NamedImage(
        title=_(u'Image'),
        required=False
    )

    attachment_1 = NamedFile(
        title=_(u'Attachment 1'),
        required=False
    )

    attachment_2 = NamedFile(
        title=_(u'Attachment 2'),
        required=False
    )

    searchable('locality')
    locality = TextLine(
        title=_(u'Locality'),
        required=False
    )

    searchable('street')
    street = TextLine(
        title=_(u'Street'),
        required=False
    )

    searchable('housenumber')
    housenumber = TextLine(
        title=_(u'Housenumber'),
        required=False
    )

    searchable('zipcode')
    zipcode = TextLine(
        title=_(u'Zipcode'),
        required=False
    )

    searchable('town')
    town = TextLine(
        title=_(u'Town'),
        required=False
    )

    searchable('location_url')
    location_url = Website(
        title=_(u'Location Website'),
        required=False
    )

    searchable('event_url')
    event_url = Website(
        title=_(u'Event Website'),
        required=False
    )

    searchable('organizer')
    organizer = TextLine(
        title=_(u'Organizer'),
        required=False
    )

    searchable('contact_name')
    contact_name = TextLine(
        title=_(u'Contact Name'),
        required=False
    )

    searchable('contact_email')
    contact_email = Email(
        title=_(u'Contact Email'),
        required=False
    )

    searchable('contact_phone')
    contact_phone = TextLine(
        title=_(u'Contact Phone'),
        required=False
    )

    searchable('prices')
    prices = Text(
        title=_(u'Prices'),
        required=False
    )

    searchable('registration')
    registration = Website(
        title=_(u'Ticket / Registration Website'),
        required=False
    )


# don't show these fields as they are not used
IEventsDirectoryItem.setTaggedValue(
    'seantis.dir.base.omitted', ['cat3', 'cat4', 'description']
)


class IEventSubmissionData(form.Schema):
    """ The event submission form uses this interfaces to get the event dates
    from the user, since the default IEventBasic interface turned out to be
    too confusing for many.

    """

    submission_date_type = List(
        title=_(u'When will this event occur?'),
        required=True,
        value_type=Choice(
            vocabulary=SimpleVocabulary([
                SimpleTerm(value='date', title=_(u'On a single day')),
                SimpleTerm(value='range', title=_(u'On several days'))
            ])
        ),
        default=['date']
    )

    submission_date = Date(
        title=_(u'Date'),
        required=False
    )

    submission_start_time = Time(
        title=_(u'Start time'),
        required=False
    )

    submission_end_time = Time(
        title=_(u'End time'),
        required=False
    )

    submission_range_start_date = Date(
        title=_(u'Start date'),
        required=False
    )

    submission_range_end_date = Date(
        title=_(u'End date'),
        required=False
    )

    submission_range_start_time = Time(
        title=_(u'Starts each day at'),
        required=False
    )

    submission_range_end_time = Time(
        title=_(u'Ends each day at'),
        required=False
    )

    submission_whole_day = Bool(
        title=_(u'All day'),
        required=False
    )

    submission_recurrence = Text(
        title=_(u'Recurrence'),
        required=False
    )

    submission_days = List(
        title=_(u'On these days'),
        value_type=Choice(vocabulary=days),
        required=False,
        default=days.keys
    )

    form.widget(
        'submission_recurrence', RecurrenceWidget,
        start_field='IEventSubmissionData.submission_date',
        first_day=first_weekday_sun0
    )


alsoProvides(IEventSubmissionData, IFormFieldProvider)


class IExternalEventSource(form.Schema):
    """Marker for external event sources (content type)"""

    title = TextLine(
        title=_(u'Name'),
    )

    enabled = Bool(
        title=_(u'Enabled'),
        required=True,
        default=True
    )

    limit = Int(
        title=_(u'Limit'),
        description=_(u'The number of events to import every 15 minutes.'),
        required=True,
        default=25
    )

    autoremove = Bool(
        title=_(u'Synchronise deleted events'),
        description=_(
            u'Automatically delete already imported events if no longer '
            u'present.'
        ),
        required=True,
        default=True
    )


class IExternalEventSourceGuidle(IExternalEventSource):
    """Guidle source."""

    url = Website(
        title=_(u'Guidle URL to download from'),
        required=True
    )


class IGuidleClassifier(Interface):
    """Interface for guidle import classifier (adapter).

        An example classifier might look something like:

        class MyGuidleClassfier(grok.Adapter):
            grok.context(EventsSourceGuidle)
            grok.provides(IGuidleClassifier)

            def classify(self, classifications):
                categories = set()
                for classification in classifications.iterchildren():
                    categories.add(classification.attrib['name'])
                return categories
    """

    def classify(self, classifications):
        """ Returns a set of categories of a guidle event. """


class IExternalEventSourceSeantisJson(IExternalEventSource):
    """Seantis events directory source."""

    url = Website(
        title=_(u'Events directory URL'),
        required=True
    )

    do_filter = Bool(
        title=_(u'Filter events'),
        required=False,
        default=False
    )

    cat1 = TextLine(
        title=_(u'What'),
        required=False,
        default=u''
    )

    cat2 = TextLine(
        title=_(u'Where'),
        required=False,
        default=u''
    )


class NoImportDataException(Exception):
    pass


class IExternalEventCollector(Interface):
    """Interface for exernal event sources (adapter)"""

    autoremove = Attribute("""
        Automatically remove already imported events which have been deleted
        externally.
        """)

    def fetch(self):
        """Generator function returning all items to import. If returning an
        empty array, all already imported events will be deleted if autoremove
        is set. To prevent importing anything, throw a NoImportDataException.

        """


class ISourceCondition(Interface):
    """Interface for the configurable aspects of a source condition of a
    content rule.

    This is also used to create add and edit forms, below.
    """

    source = TextLine(
        title=_(u'Source'),
        description=_(u"The source id to check for. "
                      u"Leave empty for any source."),
        required=False
    )


class IResourceViewedEvent(Interface):
    """ Event triggered when a seantis.events resource is viewed. Pretty
    useful if you need a hook which is guaranteed to be triggered on a plone
    site where seantis.events is active.

    """

    context = Attribute("The IResourceBase context object")


# force the user to select at least one value for each category
@form.validator(field=IDirectoryItemCategories['cat1'])
@form.validator(field=IDirectoryItemCategories['cat2'])
def validate_category(value):
    if not value:
        raise Invalid(_(u'Please choose at least one category'))


# to enforce the last rule, categories must exist
@form.validator(field=IEventsDirectory['cat1_suggestions'])
@form.validator(field=IEventsDirectory['cat2_suggestions'])
def validate_suggestion(value):
    if not value:
        raise Invalid(_(u'Please enter at least one suggestion'))


# images and attachments are limited in size
def check_filesize(value, size_in_mb, type):

    if value.getSize() > size_in_mb * 1024 ** 2:
        raise Invalid(
            _(
                u'${type} bigger than ${max} Megabyte are not allowed',
                mapping={'max': size_in_mb, 'type': type}
            )
        )


# Ensure that the uploaded image at least has an image header, a check
# which is important because users can upload files anonymously
@form.validator(field=IEventsDirectoryItem['image'])
def validate_image(value):
    if not value:
        return

    if not imghdr.what(value.filename, value.data):
        raise Invalid(_(u'Unknown image format'))

    check_filesize(value, 1, _(u'Images'))


# Attachments are limited to certain filetypes
mime_whitelist = {
    'application/pdf': _(u'PDF'),
}


@form.validator(field=IEventsDirectoryItem['attachment_1'])
@form.validator(field=IEventsDirectoryItem['attachment_2'])
def validate_attachment(value):
    if not value:
        return

    filetype = magic.from_buffer(value.data[:1024], mime=True)

    if not filetype in mime_whitelist:
        print filetype
        raise Invalid(
            _(
                u'Unsupported fileformat. Supported is ${formats}',
                mapping={'formats': u','.join(sorted(mime_whitelist.values()))}
            )
        )

    check_filesize(value, 10, _(u'Attachments'))


class ITerms(form.Schema):
    agreed = Bool(
        required=True, default=False
    )


@form.validator(field=ITerms['agreed'])
def validate_terms_and_conditions(agreed):
    if not agreed:
        raise Invalid(
            _(
                u'You have to agree to the terms '
                u'and conditions to submit this event'
            )
        )



import json
import pytz

from logging import getLogger
log = getLogger('seantis.dir.events')

from dateutil.parser import parse
from five import grok
from urllib import quote_plus
from urllib2 import urlopen

from seantis.dir.events.dates import default_now, as_timezone
from seantis.dir.events.interfaces import (
    IExternalEventCollector,
    IExternalEventSourceSeantisJson,
    NoImportDataException
)


def fix_tzinfo(date):
    """ Fix timezone information for dates parsed with dateutil."""
    if date.tzinfo and date.tzname():
        timezone = date.tzname()
        date.replace(tzinfo=None)
        date = as_timezone(date, timezone)
    return date


class EventsSourceSeantisJson(grok.Adapter):
    grok.context(IExternalEventSourceSeantisJson)
    grok.provides(IExternalEventCollector)

    def build_url(self):

        url = self.context.url.strip() + '?'
        url += 'type=json&compact=true'
        if self.context.do_filter and (self.context.cat1 or self.context.cat2):
            url += '&filter=true'
            if self.context.cat1:
                cat = quote_plus(
                    self.context.cat1.strip().encode('utf-8')
                )
                url += '&cat1=' + cat
            if self.context.cat2:
                cat = quote_plus(
                    self.context.cat2.strip().encode('utf-8')
                )
                url += '&cat2=' + cat
        return url

    def fetch(self, json_string=None):

        try:
            if json_string is None:
                url = self.build_url()
                json_string = urlopen(url, timeout=60).read()

            events = json.loads(json_string)
        except:
            raise NoImportDataException()

        for event in events:

            cat1, cat2 = event.get('cat1'), event.get('cat2')
            cat1 = set(cat1) if cat1 is not None else set()
            cat2 = set(cat2) if cat2 is not None else set()

            if self.context.do_filter:
                if self.context.cat1:
                    if self.context.cat1:
                        if self.context.cat1 not in cat1:
                            continue
                    else:
                        continue
                if self.context.cat2:
                    if self.context.cat2:
                        if self.context.cat2 not in cat2:
                            continue
                    else:
                        continue

            e = {}
            e['fetch_id'] = self.context.url
            updated = event.get('last_update')
            e['last_update'] = parse(updated) if updated else default_now()
            e['source_id'] = event['id']

            e['id'] = event.get('id')
            e['title'] = event.get('title')
            e['short_description'] = event.get('short_description')
            e['long_description'] = event.get('long_description')
            e['cat1'] = cat1
            e['cat2'] = cat2

            timezone = event.get('timezone')
            start = fix_tzinfo(parse(event.get('start')))
            end = fix_tzinfo(parse(event.get('end')))

            e['timezone'] = timezone
            e['start'] = as_timezone(start, timezone)
            e['end'] = as_timezone(end, timezone)
            e['recurrence'] = event.get('recurrence')
            e['whole_day'] = event.get('whole_day')

            e['locality'] = event.get('locality')
            e['street'] = event.get('street')
            e['housenumber'] = event.get('housenumber')
            e['zipcode'] = event.get('zipcode')
            e['town'] = event.get('town')
            e['location_url'] = event.get('location_url')
            lon, lat = event.get('longitude'), event.get('latitude')
            e['longitude'] = str(lon) if lon is not None else None
            e['latitude'] = str(lat) if lon is not None else None
            e['organizer'] = event.get('organizer')
            e['contact_name'] = event.get('contact_name')
            e['contact_email'] = event.get('contact_email')
            e['contact_phone'] = event.get('contact_phone')
            e['prices'] = event.get('prices')
            e['event_url'] = event.get('event_url')
            e['registration'] = event.get('registration')
            e['submitter'] = event.get('submitter')
            e['submitter_email'] = event.get('submitter_email')

            try:
                e['image'] = event['images'][0]['url']
                e['image_name'] = event['images'][0]['name']
            except (TypeError, KeyError, IndexError):
                e['image'] = None

            try:
                e['attachment_1'] = event['attachements'][0]['url']
                e['attachment_1_name'] = event['attachements'][0]['name']
            except (TypeError, KeyError, IndexError):
                e['attachment_1'] = None
            try:
                e['attachment_2'] = event['attachements'][1]['url']
                e['attachment_2_name'] = event['attachements'][1]['name']
            except (TypeError, KeyError, IndexError):
                e['attachment_2'] = None

            yield e



from datetime import datetime, time
from plone.app.event.dx.behaviors import IEventBasic, IEventRecurrence
from z3c.form.interfaces import ActionExecutionError
from zope.annotation.interfaces import IAnnotations
from zope.interface import implements, Invalid
from seantis.dir.events.interfaces import IEventSubmissionData
from seantis.dir.events import dates
from seantis.dir.events import _
from seantis.dir.events.recurrence import occurrences_over_limit


ANNOTATION_KEY = 'seantis.dir.events.submission-data'


def validate_event_submission(data):

    def fail(msg):
        raise ActionExecutionError(Invalid(msg))

    if data['submission_date_type'] == ['date']:
        if data.get('submission_date') is None:
            fail(_(u'Missing start date'))

        if not data.get('submission_whole_day', False):
            if data.get('submission_start_time') is None:
                fail(_(u'Missing start time'))
            if data.get('submission_end_time') is None:
                fail(_(u'Missing end time'))

    if data['submission_date_type'] == ['range']:
        if data.get('submission_range_start_date') is None:
            fail(_(u'Missing start date'))
        if data.get('submission_range_end_date') is None:
            fail(_(u'Missing end date'))

        if not data.get('submission_whole_day', False):
            if data.get('submission_range_start_time') is None:
                fail(_(u'Missing start time'))
            if data.get('submission_range_end_time') is None:
                fail(_(u'Missing end time'))

    start, end, whole_day, recurrence = get_event_dates_from_submission(data)

    if not start:
        fail(_(u'Missing start date'))

    if not end:
        fail(_(u'Missing end date'))

    if end < start:
        fail(_(u'Start date after end date'))

    # ensure that the recurrences are not over limit
    if recurrence:

        limit = 365  # one event each day for a whole year

        if occurrences_over_limit(recurrence, start, limit):
            fail(
                _(
                    u'You may not add more than ${max} occurences',
                    mapping={'max': limit}
                )
            )


def get_event_dates_from_submission(data, timezone=None):

    timezone = timezone or dates.default_timezone()
    in_timezone = lambda date: dates.as_timezone(date, timezone)

    single_day = data['submission_date_type'] == ['date']

    if single_day:
        date, start, end, whole_day, recurrence = (
            data['submission_date'],
            data['submission_start_time'],
            data['submission_end_time'],
            data['submission_whole_day'],
            data['submission_recurrence']
        )

        if whole_day:
            start = time(0, 0, 0)
            end = time(23, 59, 59)

        start, end = map(
            in_timezone, dates.combine_daterange(date, start, end)
        )

        return start, end, whole_day, recurrence
    else:
        start_date, end_date, start_time, end_time, whole_day = (
            data['submission_range_start_date'],
            data['submission_range_end_date'],
            data['submission_range_start_time'],
            data['submission_range_end_time'],
            data['submission_whole_day']
        )

        if whole_day:
            start_time = time(0, 0, 0)
            end_time = time(23, 59, 59)

        start, end = map(
            in_timezone, dates.combine_daterange(
                start_date, start_time, end_time
            )
        )

        recurrence = 'RRULE:FREQ=WEEKLY;UNTIL={}'.format(
            dates.as_rfc5545_string(in_timezone(
                datetime.combine(end_date, time(23, 59, 59))
            ))
        )

        if data['submission_days']:
            recurrence += ';BYDAY={}'.format(
                ','.join(
                    [str(d) for d in data['submission_days']]
                )
            )

        return start, end, whole_day, recurrence


class EventSubmissionData(object):

    implements(IEventSubmissionData)

    def __init__(self, context):
        self.context = context
        self.fields = IEventSubmissionData.names()
        self.annotation = IAnnotations(self.context)
        self.data = self.annotation.get(ANNOTATION_KEY, {})
        self.annotation[ANNOTATION_KEY] = self.data

    def __getattr__(self, name):
        if name in self.__dict__.get('fields', []):
            if name in self.data:
                return self.data[name]

        raise AttributeError

    def __setattr__(self, name, value):
        if name in self.__dict__.get('fields', []):
            self.data[name] = value
        else:
            self.__dict__[name] = value

    def inject_sane_dates(self):
        """ Takes the IEventSubmissionDate data and makes nice IEventBasic
        data out of it.

        """
        basic = IEventBasic(self.context)
        recurring = IEventRecurrence(self.context)

        (
            basic.start,
            basic.end,
            basic.whole_day,
            recurring.recurrence
        ) = get_event_dates_from_submission(self.data, basic.timezone)

        self.submission_recurrence = recurring.recurrence



# -- coding: utf-8 --

from copy import copy
from five import grok
from collections import OrderedDict
from datetime import date

from Acquisition import aq_inner, aq_base
from Acquisition.interfaces import IAcquirer

from plone.directives import form
from plone.z3cform.fieldsets import extensible
from plone.dexterity.utils import createContent, addContentToContainer
from plone.formwidget.recurrence.z3cform.widget import RecurrenceWidget

from Products.statusmessages.interfaces import IStatusMessage

import zope.event
import zope.lifecycleevent

from zope.interface import implements
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.schema import Choice, TextLine
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary
from zope.component.interfaces import ComponentLookupError

from z3c.form import field, group, button, widget
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from z3c.form.browser.radio import RadioFieldWidget

from collective.z3cform.mapwidget.widget import MapFieldWidget
from collective.geo.contentlocations.interfaces import IGeoManager

from plone.formwidget.datetime.z3cform.widget import DateWidget
from plone.app.event.base import default_timezone
from plone.app.event.dx.behaviors import first_weekday_sun0

from seantis.dir.base import utils as base_utils
from seantis.dir.base.interfaces import IDirectoryPage, IDirectoryCategorized

from seantis.dir.events.interfaces import (
    IEventsDirectory,
    IEventsDirectoryItem,
    ITerms,
    IEventSubmissionData,
    IExternalEvent
)

from seantis.dir.events.token import (
    verify_token,
    apply_token,
    clear_token,
    append_token,
    event_by_token,
    current_token
)

from seantis.dir.events import utils, dates
from seantis.dir.events.submission import validate_event_submission
from seantis.dir.events.recurrence import occurrences, grouped_occurrences
from seantis.dir.events import _


# plone.formwidget.recurrence got rid of the parameterized widget
# in favor ofusing plone.autoform. unfortunately, plone.autoform seems
# incapable of dealing with groups instead forms.
#
# I hate form libs.
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget


class ParameterizedFieldWidget(object):
    implements(IFieldWidget)

    def __new__(cls, field, request):
        widget = FieldWidget(field, cls.widget(request))
        for k, v in cls.kw.items():
            setattr(widget, k, v)
        return widget


def ParameterizedWidgetFactory(widget, **kw):
    return type('%sFactory' % widget.__name__,
                (ParameterizedFieldWidget,),
                {'widget': widget, 'kw': kw})


class EventBaseGroup(group.Group):

    group_fields = {}
    dynamic_fields = []

    _cached_fields = []

    @property
    def fields(self):
        """ Returns the fields defined in group_fields, making sure that
        dynamic fields are shallow copied before they are touched.

        Dynamic fields in this context are fields whose schame properties
        are changed on the fly and it is important that they are copied
        since these changes otherwise leak through to other forms.

        """

        if self._cached_fields:
            return self._cached_fields

        result = None

        for interface, fields in self.group_fields.items():
            if result is None:
                result = field.Fields(interface).select(*fields)
            else:
                result += field.Fields(interface).select(*fields)

        for f in self.dynamic_fields:
            if f in result:
                result[f].field = copy(result[f].field)

        self._cached_fields = result
        return result

    def updateWidgets(self):
        self.update_dynamic_fields()
        super(EventBaseGroup, self).updateWidgets()
        self.update_widgets()

    def update_dynamic_fields(self):
        """ Called when it's time to update the dynamic fields. """

    def update_widgets(self):
        """ Called when it's time to make changes to the widgets. """


class NavigationStep(object):
    def __init__(self, id=None, text=None, url=None):
        self.id, self.text, self.url = id, text, url


class NavigationMixin(object):

    _steps = None

    @property
    def show_navigation(self):
        return self.__name__ in ('submit', 'preview', 'finish')

    @property
    def steps(self):

        if self._steps:
            return self._steps

        steps = [
            NavigationStep('submit', _(u'Enter'), None),
            NavigationStep('preview', _(u'Verify'), None),
            NavigationStep('finish', _(u'Finish'), None)
        ]

        if self.__name__ == 'submit':
            steps[0].url = self.context.absolute_url() + '/@@submit'
            steps[1].url = None
            steps[2].url = None
        elif self.__name__ == 'preview':
            steps[0].url = self.directory.absolute_url() + '/@@submit'
            steps[1].url = self.context.absolute_url() + '/@@preview'
            steps[2].url = None
        elif self.__name__ == 'finish':
            steps[0].url = self.directory.absolute_url() + '/@@submit'
            steps[1].url = self.context.absolute_url() + '/@@preview'
            steps[2].url = self.context.absolute_url() + '/@@finish'

        for i in range(0, len(steps)):
            if steps[i].url:
                steps[i].url = append_token(self.context, steps[i].url)

        self._steps = steps
        return self._steps

    @property
    def current_step(self):
        for step in self.steps:
            if self.__name__ == step.id:
                return step

    def step_classes(self, step):

        classes = ["formTab"]

        if step.id == self.steps[0].id:
            classes.append("firstFormTab")

        if step.id == self.steps[-1].id:
            classes.append("lastFormTab")

        if step.id == self.current_step.id:
            classes.append("selected")

        if self.steps.index(step) < self.steps.index(self.current_step):
            classes.append("visited")

        return ' '.join(classes)


class GeneralGroup(EventBaseGroup):

    label = _(u'Event')

    dynamic_fields = ('cat1', 'cat2')

    group_fields = OrderedDict()
    group_fields[IEventsDirectoryItem] = (
        'title', 'short_description', 'long_description'
    )
    group_fields[IDirectoryCategorized] = (
        'cat1', 'cat2'
    )

    def update_dynamic_fields(self):
        categories = (self.fields['cat1'], self.fields['cat2'])

        for category in categories:
            category.field.description = u''
            category.field.value_type = Choice(
                source=self.available_categories(
                    self.context, category.__name__
                )
            )

        categories[0].widgetFactory = CheckBoxFieldWidget
        categories[0].field.required = True
        categories[1].widgetFactory = RadioFieldWidget
        categories[1].field.required = True

    def update_widgets(self):
        # update labels of categories
        labels = self.context.labels()
        widgets = [w for w in self.widgets if w in labels]

        for widget in widgets:
            self.widgets[widget].label = labels[widget]

        self.widgets['title'].label = _(u'Title of event')

    def available_categories(self, context, category):

        @grok.provider(IContextSourceBinder)
        def get_categories(ctx):
            terms = []

            for value in context.suggested_values(category):
                terms.append(
                    SimpleVocabulary.createTerm(value, hash(value), value)
                )

            return SimpleVocabulary(terms)

        return get_categories


class DateGroup(EventBaseGroup):

    label = _(u'Event date')

    dynamic_fields = (
        'submission_date',
        'submission_range_start_date',
        'submission_range_end_date',
        'submission_date_type',
        'submission_recurrence',
        'submission_days'
    )

    group_fields = OrderedDict()

    group_fields[IEventSubmissionData] = (
        'submission_date_type',
        'submission_date',
        'submission_start_time',
        'submission_end_time',
        'submission_range_start_date',
        'submission_range_end_date',
        'submission_range_start_time',
        'submission_range_end_time',
        'submission_whole_day',
        'submission_recurrence',
        'submission_days'
    )

    def update_dynamic_fields(self):
        recurrence = self.fields['submission_recurrence']
        recurrence.widgetFactory = ParameterizedWidgetFactory(
            RecurrenceWidget, start_field='submission_date'
        )

        self.fields['submission_date_type'].widgetFactory = RadioFieldWidget

        # plone.formwidget.recurrencewidget needs this... obviously?
        date_widget_fields = (
            'submission_date',
            'submission_range_start_date',
            'submission_range_end_date'
        )
        for field in date_widget_fields:
            self.fields[field].widgetFactory = ParameterizedWidgetFactory(
                DateWidget, first_day=first_weekday_sun0
            )

        self.fields['submission_days'].widgetFactory = CheckBoxFieldWidget

        default_date = date.today()
        self.fields['submission_date'].field.default = default_date
        self.fields['submission_range_start_date'].field.default = default_date
        self.fields['submission_range_end_date'].field.default = default_date


class LocationGroup(EventBaseGroup):

    label = _(u'Location')

    group_fields = OrderedDict()
    group_fields[IEventsDirectoryItem] = (
        'locality', 'street', 'housenumber', 'zipcode', 'town', 'location_url'
    )


class MapGroup(EventBaseGroup):

    label = _(u'Map')

    dynamic_fields = ('wkt', )

    group_fields = OrderedDict()
    group_fields[IGeoManager] = (
        'wkt',
    )

    def update_widgets(self):
        self.widgets['wkt'].label = _(u'Coordinates')

    def update_dynamic_fields(self):
        coordinates = self.fields['wkt']
        coordinates.widgetFactory = MapFieldWidget


class InformationGroup(EventBaseGroup):
    label = _(u'Information')
    fields = field.Fields(IEventsDirectoryItem).select(
        'organizer', 'contact_name', 'contact_email',
        'contact_phone', 'prices', 'event_url', 'registration',
        'image', 'attachment_1', 'attachment_2'
    )


class EventSubmissionForm(extensible.ExtensibleForm):

    grok.baseclass()
    grok.require('zope2.View')

    coordinates = None

    def directory(self):
        return self.context

    def prepare_coordinates(self, data):
        if data.get('wkt'):
            self.coordinates = utils.verify_wkt(data['wkt']).__geo_interface__
        else:
            self.coordinates = None

        del data['wkt']

    def apply_coordinates(self, content):
        c = self.coordinates
        if c:
            IGeoManager(content).setCoordinates(c['type'], c['coordinates'])
        else:
            IGeoManager(content).removeCoordinates()

    def handle_cancel(self):
        try:
            clear_token(self.context)
        except ComponentLookupError:
            pass
        IStatusMessage(self.request).add(
            _(u"Event submission cancelled"), "info"
        )
        self.request.response.redirect(self.directory.absolute_url())


class EventSubmitForm(extensible.ExtensibleForm, form.Form, NavigationMixin):
    """Event submission form mainly targeted at anonymous users.

    This form combines add- and edit-form functionality to streamline
    the user's experience. The idea is that this form and the
    PreviewForm below work in tandem:

    1. User opens directory/submit and enters an event (add-form)
    2. User saves result and is directed to directory/event/preview
    3. User changes his mind and clicks the back button
    4. User resubmits the form (which is now an edit-form)

    This can go back and forth providing a tight feedback-loop.
    If an edit form is thrown into the mix, the back button cannot
    work correctly because it will lead the user to the add form
    even if he really means to edit.

    The form is an edit-form if the user has a token in the session which
    machtes an existing event in preview state.

    The form is an add-form if the user does not have a token or it does
    not exist in the database.

    If the user clicks 'cancel' the token is thrown away, leaving
    an event in limbo. This will be cleaned up by a cron job.

    """

    implements(IDirectoryPage)

    grok.name('submit')
    grok.require('cmf.RequestReview')
    grok.context(IEventsDirectory)

    template = ViewPageTemplateFile('templates/form.pt')

    portal_type = 'seantis.dir.events.item'

    groups = (
        GeneralGroup, DateGroup, LocationGroup, MapGroup, InformationGroup
    )
    enable_form_tabbing = False

    label = _(u'Event Submission Form')
    description = _(u'Send us your events and we will publish them')

    # if true, render() will return nothing
    empty_body = False

    def __init__(self, context, request, fti=None):
        super(EventSubmitForm, self).__init__(context, request)

        # do not show the edit/action bar
        self.request['disable_border'] = True

    def __of__(self, context):
        return self

    def update(self):
        self.setup_form()
        super(EventSubmitForm, self).update()

    def render(self):
        if self.empty_body:
            return ''

        return super(EventSubmitForm, self).render()

    def redirect(self, url):
        self.empty_body = True
        self.request.response.redirect(url)

    def message(self, message, type="info"):
        IStatusMessage(self.request).add(message, type)

    @property
    def directory(self):
        return self.context

    def form_type(self):
        self.event = event_by_token(
            self.directory, current_token(self.request)
        )
        return self.event and 'editform' or 'addform'

    def setup_form(self):
        self.buttons = button.Buttons()
        self.handlers = button.Handlers()

        if self.form_type() == 'addform':
            preview = button.Button(title=_(u'Continue'), name='save')
            self.buttons += button.Buttons(preview)

            preview_handler = button.Handler(
                preview, self.__class__.handle_preview
            )
            self.handlers.addHandler(preview, preview_handler)

            self.ignoreContext = True
            self.ignoreReadonly = True
        else:
            update = button.Button(title=_(u'Continue'), name='save')
            self.buttons += button.Buttons(update)

            update_handler = button.Handler(
                update, self.__class__.handle_update
            )
            self.handlers.addHandler(update, update_handler)

            self.context = self.event

        cancel = button.Button(title=_(u'Cancel'), name='cancel')
        self.buttons += button.Buttons(cancel)

        cancel_handler = button.Handler(cancel, self.__class__.handle_cancel)
        self.handlers.addHandler(cancel, cancel_handler)

    def prepare_coordinates(self, data):
        """ The coordinates need to be verified and converted. First they are
        converted and kept out of the item creation process. Later they
        are added to the object once that exists.

        """
        if data.get('wkt'):
            self.coordinates = utils.verify_wkt(data['wkt']).__geo_interface__
        else:
            self.coordinates = None

        del data['wkt']

    def apply_coordinates(self, content):
        c = self.coordinates
        if c:
            IGeoManager(content).setCoordinates(c['type'], c['coordinates'])
        else:
            IGeoManager(content).removeCoordinates()

    def prepare_submission(self, data):
        self.submission = {}

        submission_fields = IEventSubmissionData.names()
        submitted_fields = data.keys()

        for field in submitted_fields:
            if field in submission_fields:
                self.submission[field] = data[field]
                del data[field]

    def apply_submission(self, content):
        submission = IEventSubmissionData(content)

        for field, value in self.submission.items():
            setattr(submission, field, value)

        submission.inject_sane_dates()

    def handle_preview(self, action):
        data, errors = self.extractData()
        validate_event_submission(data)

        if errors:
            self.status = self.formErrorsMessage
            return

        obj = self.create_and_add(data)
        if obj is not None:
            url = self.context.absolute_url() + '/' + obj.id + '/preview'
            self.redirect(append_token(obj, url))

    def handle_update(self, action):
        data, errors = self.extractData()
        validate_event_submission(data)

        if errors:
            self.status = self.formErrorsMessage
            return

        self.prepare_coordinates(data)
        self.apply_coordinates(self.getContent())

        self.prepare_submission(data)
        self.apply_submission(self.getContent())

        changes = self.applyChanges(data)

        if changes:
            self.message(_(u'Event Preview Updated'))
        else:
            self.message(_(u'No changes were applied'))

        url = self.context.absolute_url() + '/preview'
        self.redirect(append_token(self.context, url))

    def handle_cancel(self, action):
        try:
            clear_token(self.context)
        except ComponentLookupError:
            pass

        self.message(_(u"Event submission cancelled"))
        self.redirect(self.directory.absolute_url())

    def create(self, data):
        data['timezone'] = default_timezone()

        self.prepare_coordinates(data)
        self.prepare_submission(data)

        content = createContent('seantis.dir.events.item', **data)

        if IAcquirer.providedBy(content):
            content = content.__of__(aq_inner(self.context))

        # must be done before adding to the container
        self.apply_submission(content)

        return aq_base(content)

    def add(self, obj):
        # not checking the contrains means two things
        # * impossible content types could theoretically added
        # * anonymous users can post events
        self.content = addContentToContainer(
            aq_inner(self.context), obj, checkConstraints=False
        )

        self.apply_coordinates(self.content)
        apply_token(self.content)

    def create_and_add(self, data):
        obj = self.create(data)
        zope.event.notify(zope.lifecycleevent.ObjectCreatedEvent(obj))
        self.add(obj)
        return obj

    @property
    def allow_edit(self):
        return True

    @property
    def import_source(self):
        return ""


class EventEditForm(EventSubmitForm):

    implements(IDirectoryPage)

    grok.name('edit')
    grok.require('cmf.ModifyPortalContent')
    grok.context(IEventsDirectoryItem)

    @property
    def allow_edit(self):
        return not IExternalEvent.providedBy(self.context)

    @property
    def import_source(self):
        try:
            result = IExternalEvent(self.context).source
        except:
            result = ""
        return result

    def setup_form(self):
        self.buttons = button.Buttons()
        self.handlers = button.Handlers()

        save = button.Button(title=_(u'Save Event'), name='save')
        self.buttons += button.Buttons(save)

        save_handler = button.Handler(save, self.__class__.handle_save)
        self.handlers.addHandler(save, save_handler)

        self.event = self.context

        cancel = button.Button(
            title=_(u'Cancel Event Submission'), name='cancel'
        )
        self.buttons += button.Buttons(cancel)

        cancel_handler = button.Handler(cancel, self.__class__.handle_cancel)
        self.handlers.addHandler(cancel, cancel_handler)

    def handle_save(self, action):
        if not self.allow_edit:
            self.message(_(u'Imported events may not be edited, '
                           u'no changes where applied'))
            return

        data, errors = self.extractData()
        validate_event_submission(data)

        if errors:
            self.status = self.formErrorsMessage
            return

        self.prepare_coordinates(data)
        self.apply_coordinates(self.getContent())

        self.prepare_submission(data)
        self.apply_submission(self.getContent())

        changes = self.applyChanges(data)

        if changes:
            self.message(_(u'Event Saved'))
        else:
            self.message(_(u'No changes were applied'))

        url = self.context.absolute_url()
        self.redirect(append_token(self.context, url))


class IPreview(form.Schema):
    title = TextLine(required=False)


class DetailPreviewWidget(widget.Widget):

    _template = ViewPageTemplateFile('templates/previewdetail.pt')

    def render(self):
        self.directory = self.context.get_parent()
        return self._template(self)

    def safe_html(self, html):
        return base_utils.safe_html(html)


class ListPreviewWidget(DetailPreviewWidget):

    _template = ViewPageTemplateFile('templates/previewlist.pt')
    show_map = False

    @base_utils.cached_property
    def occurrence_groups(self):
        dr = dates.DateRanges()
        start, end = dr.this_year[0], dr.next_year[1]
        result = grouped_occurrences(
            occurrences(self.context, start, end), self.request
        )
        return result

    @base_utils.cached_property
    def occurrences_count(self):
        count = 0

        for day, occurrences in self.occurrence_groups.items():
            count += len(occurrences)

        return count


def DetailPreviewFieldWidget(field, request):
    """IFieldWidget factory for MapWidget."""
    return widget.FieldWidget(field, DetailPreviewWidget(request))


def ListPreviewFieldWidget(field, request):
    """IFieldWidget factory for MapWidget."""
    return widget.FieldWidget(field, ListPreviewWidget(request))


class PreviewGroup(EventBaseGroup):

    label = _(u'Detail Preview')

    dynamic_fields = ('title', )

    group_fields = OrderedDict()
    group_fields[IPreview] = ('title', )

    def update_dynamic_fields(self):
        self.fields['title'].widgetFactory = DetailPreviewFieldWidget


class ListPreviewGroup(PreviewGroup):

    label = _(u'List Preview')

    def update_dynamic_fields(self):
        self.fields['title'].widgetFactory = ListPreviewFieldWidget

    def update_widgets(self):
        occurrences = self.widgets['title'].occurrences_count
        if occurrences > 1:
            self.label = _(u'List Preview (${number} Occurrences)', mapping={
                'number': occurrences
            })
        else:
            self.label = _(u'List Preview (No Occurrences)')


class SubmitterGroup(EventBaseGroup):

    label = _(u'Submitter')

    dynamic_fields = ('submitter', 'submitter_email')

    group_fields = OrderedDict()
    group_fields[IEventsDirectoryItem] = ('submitter', 'submitter_email')
    group_fields[ITerms] = ('agreed', )

    def update_dynamic_fields(self):
        self.fields['submitter'].field.required = True
        self.fields['submitter_email'].field.required = True

        # remove the terms and conditions agreement if there is none
        if not self.context.get_parent().terms:
            del self.fields['agreed']
        else:
            # otherwise be sure to link to it
            url = self.context.get_parent().absolute_url() + '/@@terms'
            self.fields['agreed'].field.description = utils.translate(
                self.request, _(
                    u"I agree to the <a target='_blank' href='${url}'>"
                    u"Terms and Conditions</a>",
                    mapping={'url': url}
                )
            )


class PreviewForm(EventSubmissionForm, form.AddForm, NavigationMixin):

    implements(IDirectoryPage)

    grok.context(IEventsDirectoryItem)
    grok.name('preview')

    groups = (PreviewGroup, ListPreviewGroup)
    template = ViewPageTemplateFile('templates/previewform.pt')

    enable_form_tabbing = True

    label = _(u'Event Submission Preview')
    description = u''

    @property
    def directory(self):
        return self.context.get_parent()

    def current_token(self):
        return current_token(self.request)

    def update(self, *args, **kwargs):
        verify_token(self.context, self.request)
        super(PreviewForm, self).update(*args, **kwargs)

    @button.buttonAndHandler(_('Continue'), name='save')
    def handleSubmit(self, action):
        self.request.response.redirect(
            append_token(
                self.context, self.context.absolute_url() + '/@@finish'
            )
        )

    @button.buttonAndHandler(_(u'Adjust'), name='adjust')
    def handleAdjust(self, action):
        self.request.response.redirect(
            append_token(
                self.context, self.directory.absolute_url() + '/@@submit'
            )
        )

    @button.buttonAndHandler(_(u'Cancel'), name='cancel')
    def handleCancel(self, action):
        self.handle_cancel()


class FinishForm(EventSubmissionForm, form.AddForm, NavigationMixin):

    implements(IDirectoryPage)

    grok.context(IEventsDirectoryItem)
    grok.name('finish')

    groups = (SubmitterGroup, )

    template = ViewPageTemplateFile('templates/finishform.pt')

    enable_form_tabbing = False

    label = _(u'Event Submission Finish')
    description = u''

    @property
    def directory(self):
        return self.context.get_parent()

    def current_token(self):
        return current_token(self.request)

    def update(self, *args, **kwargs):
        verify_token(self.context, self.request)
        super(FinishForm, self).update(*args, **kwargs)

    @button.buttonAndHandler(_('Submit'), name='save')
    def handleSubmit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        self.context.submitter = data['submitter']
        self.context.submitter_email = data['submitter_email']

        clear_token(self.context)
        self.context.submit()
        IStatusMessage(self.request).add(_(u"Event submitted"), "info")
        self.request.response.redirect(self.directory.absolute_url())

    @button.buttonAndHandler(_(u'Back'), name='back')
    def handleBack(self, action):
        self.request.response.redirect(
            append_token(
                self.context, self.context.absolute_url() + '/@@preview'
            )
        )

    @button.buttonAndHandler(_(u'Cancel'), name='cancel')
    def handleCancel(self, action):
        self.handle_cancel()



import transaction

from plone.testing import z2
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting

from Testing import ZopeTestCase as ztc
from OFS.Folder import Folder


class Fixture(PloneSandboxLayer):

    default_bases = (PLONE_FIXTURE, )

    class Session(dict):
        def set(self, key, value):
            self[key] = value

    def setUpZope(self, app, configurationContext):

        import seantis.dir.events
        self.loadZCML(package=seantis.dir.events)

        app.REQUEST['SESSION'] = self.Session()

        if not hasattr(app, 'temp_folder'):
            app._setObject('temp_folder', Folder('temp_folder'))
            transaction.commit()
            ztc.utils.setupCoreSessions(app)

        # needed by plone.app.event
        z2.installProduct(app, 'Products.DateRecurringIndex')

    def tearDownZope(self, app):
        z2.uninstallProduct(app, 'Products.DateRecurringIndex')

    def setUpPloneSite(self, portal):
        self.applyProfile(portal, 'seantis.dir.events:default')


FIXTURE = Fixture()
INTEGRATION_TESTING = IntegrationTesting(
    bases=(FIXTURE,),
    name='seantis.dir.events:Integration'
)
FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(FIXTURE,),
    name='seantis.dir.events:Functional'
)



import os

import unittest2 as unittest

from datetime import datetime, timedelta

from AccessControl import getSecurityManager
from plone.testing import z2
from plone.dexterity.utils import createContentInContainer
from plone.app import testing
from Products.CMFCore.utils import getToolByName
from Products.CMFCore import permissions
from collective.betterbrowser import new_browser

from zope.component import getAdapter
from zope.component.hooks import getSite

from seantis.dir.base.interfaces import IDirectoryCatalog
from seantis.dir.events.tests.layer import INTEGRATION_TESTING
from seantis.dir.events.tests.layer import FUNCTIONAL_TESTING
from seantis.dir.events.maintenance import clear_clockservers


class IntegrationTestCase(unittest.TestCase):

    layer = INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        # create a default directory for events
        self.login_admin()
        self.directory = createContentInContainer(
            getSite(), 'seantis.dir.events.directory'
        )
        self.catalog = getAdapter(self.directory, IDirectoryCatalog)

        # Add missing indexes
        if 'source' not in self.catalog.catalog.indexes():
            self.catalog.catalog.addIndex('source', 'FieldIndex')
            self.catalog.catalog.manage_reindexIndex(ids=['source'])

        # Add environment variables
        os.environ['seantis_events_import'] = 'true'

        self.directory.manage_setLocalRoles(testing.TEST_USER_ID, ('Manager',))
        self.logout()

    def tearDown(self):
        clear_clockservers()
        testing.logout()

    def login_admin(self):
        """ Login as site owner (does not work with testing.login)"""
        z2.login(self.app['acl_users'], 'admin')

    def login_testuser(self):
        """ Login as test-user (does not work with z2.login)"""
        testing.login(self.portal, 'test-user')

    def logout(self):
        testing.logout()

    @property
    def workflow(self):
        return getToolByName(self.portal, "portal_workflow")

    def do_action(self, obj, action):
        self.workflow.doActionFor(obj, action)

    def create_event(self, **kw):
        """ Create an event in self.directory. By default, test-user must
        be logged in or the creation will be unauthorized.

        """

        defaults = {
            'start': datetime.today(),
            'end': datetime.today() + timedelta(seconds=60 * 60),
            'timezone': 'Europe/Zurich',
            'recurrence': ''
        }

        for attr in defaults:
            if not attr in kw:
                kw[attr] = defaults[attr]

        return createContentInContainer(
            self.directory, 'seantis.dir.events.item', **kw
        )

    def has_permission(self, obj, permission):
        return bool(getSecurityManager().checkPermission(permission, obj))

    def may_view(self, obj):
        return self.has_permission(obj, permissions.View)


class FunctionalTestCase(IntegrationTestCase):

    layer = FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        # Add environment variables
        os.environ['seantis_events_import'] = 'true'

    def new_browser(self):
        return new_browser(self.layer)

    def tearDown(self):
        pass


class BrowserTestCase(FunctionalTestCase):

    def setUp(self):
        super(BrowserTestCase, self).setUp()

        self.baseurl = self.portal.absolute_url()

        browser = self.new_browser()
        browser.login_admin()

        browser.open(
            '/portal_catalog/manage_addProduct/PluginIndexes/addFieldIndex'
        )
        browser.getControl(name='id').value = 'source'
        browser.getControl(
            name='extra.indexed_attrs:record:string'
        ).value = 'source'
        browser.getControl('Add').click()
        self.assertTrue('source' in browser.contents)

        # create an events directory
        browser.open('/++add++seantis.dir.events.directory')

        browser.getControl(name='form.widgets.title').value = 'Veranstaltungen'
        browser.getControl(
            name='form.widgets.cat1_suggestions'
        ).value = "Category1\nCategory1_2"
        browser.getControl(
            name='form.widgets.cat2_suggestions'
        ).value = "Category2\nCategory2_2"
        browser.getControl('Save').click()

        self.assertTrue('Veranstaltungen' in browser.contents)

        # the directory needs to be published for the anonymous
        # user to submit events
        browser.open(
            browser.url + '/../content_status_modify?workflow_action=publish'
        )

        self.admin_browser = browser

    def tearDown(self):
        self.admin_browser.open('/veranstaltungen/delete_confirmation')
        self.admin_browser.getControl('Delete').click()
        self.admin_browser.assert_notfound('/veranstaltungen')

    def addEvent(self, title='title', description='description',
                 cat1='Category1', cat2='Category2',
                 whole_day=False,
                 date=datetime.today(), start='2:00 PM', end='4:00 PM',
                 submitter='submitter', email='submitter@example.com',
                 do_submit=True, check_submitted=True,
                 do_publish=True, check_published=True):

        browser = self.admin_browser

        # Add form
        browser.open('/veranstaltungen/++add++seantis.dir.events.item')
        # browser.show()
        browser.widget('title').value = title
        browser.widget('short_description').value = description
        browser.getControl(cat1).selected = True
        browser.getControl(cat2).selected = True
        browser.widget('submission_date_type').value = ['date']
        browser.set_date('submission_date', date)
        browser.widget('submission_recurrence').value = ''
        if whole_day:
            browser.getControl('All day').selected = True
        else:
            browser.widget('submission_start_time').value = start
            browser.widget('submission_end_time').value = end
        browser.getControl('Continue').click()

        # Preview
        browser.getControl('Continue').click()

        if do_submit:

            # Submit
            browser.getControl(name='form.widgets.submitter').value = submitter
            browser.getControl(
                name='form.widgets.submitter_email').value = email
            browser.getControl('Submit').click()

            # Check if submitted
            if check_submitted:
                browser.open('/veranstaltungen?state=submitted')
                self.assertTrue(title in browser.contents)
                self.assertTrue(description in browser.contents)
                self.assertTrue(cat1 in browser.contents)
                self.assertTrue(cat2 in browser.contents)

            if do_publish:

                # Publish
                browser.open('/veranstaltungen?state=submitted')
                browser.getLink('Publish', index=1).click()

                # Check if published
                if check_published:
                    browser.open('/veranstaltungen?state=published')
                    self.assertTrue(title in browser.contents)
                    self.assertTrue(description in browser.contents)
                    self.assertTrue(cat1 in browser.contents)
                    self.assertTrue(cat2 in browser.contents)



import logging
log = logging.getLogger('seantis.dir.events')

import transaction
from transaction.interfaces import ISavepointDataManager
from transaction._transaction import AbortSavepoint

from datetime import datetime, timedelta
from five import grok
from itertools import ifilter, islice
from functools import partial

from plone.app.event.ical.exporter import construct_icalendar
from plone.memoize import instance

from threading import Lock
from plone.synchronize import synchronized

from zope.interface import implements
from zope.annotation.interfaces import IAnnotations
from zope.lifecycleevent.interfaces import (
    IObjectMovedEvent,
    IObjectModifiedEvent,
    IObjectRemovedEvent
)
from Products.CMFCore.interfaces import IActionSucceededEvent

from seantis.dir.base.catalog import DirectoryCatalog, is_exact_match
from seantis.dir.base.interfaces import IDirectoryCatalog
from seantis.dir.base.utils import previous_and_next, cached_property

from seantis.dir.events import utils
from seantis.dir.events import dates
from seantis.dir.events import recurrence
from seantis.dir.events.interfaces import (
    IEventsDirectory, IEventsDirectoryItem, IExternalEvent,
    IExternalEventSource
)

from blist import sortedset


# this is just awful
# http://plone.293351.n2.nabble.com/Event-on-object-deletion-td3670562.html
# http://stackoverflow.com/questions/11218272/plone-reacting-to-object-removal
# http://repoze.org/tmdemo.html
class ReindexDataManager(object):

    implements(ISavepointDataManager)

    def __init__(self, request, directory):
        self.request = request
        self.directory = directory
        self.transaction_manager = transaction.manager

    def tpc_begin(self, transaction):
        pass

    def tpc_finish(self, transaction):
        pass

    def tpc_abort(self, transaction):
        pass

    def commit(self, transaction):
        # We need to reindex the directory during the transaction since we
        # we store the index (as text) in the annotations (which itself
        # is an attribute of the directory and therefore stored in the ZODB)
        reindex_directory(self.directory)

    def abort(self, transaction):
        pass

    def tpc_vote(self, transaction):
        pass

    def sortKey(self):
        return id(self)

    def savepoint(self):
        """ Make it possible to enter a savepoint with this manager active. """
        return AbortSavepoint(self, transaction.get())


def reindex_already_attached():
    for resource in transaction.get()._resources:
        if isinstance(resource, ReindexDataManager):
            return True

    return False


_attach_lock = Lock()


@synchronized(_attach_lock)
def attach_reindex_to_transaction(directory):
    assert directory is not None

    request = getattr(directory, 'REQUEST', None)

    if request and not reindex_already_attached():
        transaction.get().join(ReindexDataManager(request, directory))

    if not request:
        log.warn('request not found')


def may_reindex_directory(directory):
    if not directory:
        return False

    if hasattr(directory, '_v_fetching') and getattr(directory, '_v_fetching'):
        return False

    if not IEventsDirectory.providedBy(directory):
        return False

    return True


def reindex_directory(directory):
    if may_reindex_directory(directory):
        utils.get_catalog(directory).reindex()


@grok.subscribe(IEventsDirectoryItem, IObjectRemovedEvent)
def onRemovedItem(item, event):
    attach_reindex_to_transaction(event.oldParent)


@grok.subscribe(IEventsDirectoryItem, IObjectMovedEvent)
def onMovedItem(item, event):
    if event.oldParent is not None:
        attach_reindex_to_transaction(event.oldParent)
    if event.newParent is not None:
        attach_reindex_to_transaction(event.newParent)


@grok.subscribe(IEventsDirectoryItem, IObjectModifiedEvent)
@grok.subscribe(IEventsDirectoryItem, IActionSucceededEvent)
def onModifiedItem(item, event):
    attach_reindex_to_transaction(item.get_parent())


class LazyList(object):

    def __init__(self, get_item, length):
        assert callable(get_item)

        self._get_item = get_item
        self.length = length
        self.cache = [get_item] * length

    def get_item(self, index):
        if self.cache[index] == self._get_item:
            self.cache[index] = self.cache[index](index)

        return self.cache[index]

    def __len__(self):
        return self.length

    def __getitem__(self, key):
        if isinstance(key, slice):
            return map(self.get_item, range(*key.indices(self.length)))

        if (key + 1) > self.length:
            raise IndexError

        return self.get_item(key)


class EventIndex(object):

    version = "1"

    def __init__(self, catalog, initial_index=None):
        self.catalog = catalog
        self.datekey = '%Y.%m.%d'

        # be careful here, there's self.index a property which gets
        # the index from the annotation. Initial_index is for debugging
        # only and must not overwrite an existing index otherwise.
        if initial_index is not None:
            self.index = initial_index

        if not self.index:
            self.reindex()

    @property
    def name(self):
        raise NotImplementedError

    def update(self, events):
        raise NotImplementedError

    def remove(self, events):
        raise NotImplementedError

    def reindex(self, events=[]):
        raise NotImplementedError

    @property
    def annotations(self):
        return IAnnotations(self.catalog.directory)

    @property
    def index_key(self):
        return 'seantis.dir.events.' + self.name + '-' + self.version

    def meta_key(self, key):
        return self.index_key + '_meta_' + key

    def get_index(self):
        return self.annotations.get(self.index_key, None)

    def set_index(self, value):
        self.annotations[self.index_key] = value

    index = property(get_index, set_index)

    def get_metadata(self, key):
        return self.annotations.get(self.meta_key(key), None)

    def set_metadata(self, key, value):
        self.annotations[self.meta_key(key)] = value

    def spawn_events(self, real, start=None, end=None):

        if not start or not end:
            start, end = map(dates.to_utc, dates.eventrange())

        return self.catalog.spawn(real, start, end)

    def event_by_id_and_date(self, id, date):

        real = self.catalog.catalog(
            path={'query': self.catalog.path, 'depth': 1},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=self.state,
            id=id
        )[0].getObject()

        # there is currently no way to easily look up the event by date
        # if it has been split over dates already (which is what happens
        # when the events are indexed)

        # therefore we need to currently loop over all events to find the
        # right one. certainly this can be optimized.

        # however, tests on a 10k sites with 60% of all events being recurrent
        # indicate that it's not that big of a problem. spawning events is
        # quite fast and it only happens for 10 items per request

        # still, I would prefer some kind of lookup here

        min_date = date - timedelta(days=1)
        max_date = date + timedelta(days=1)

        for item in self.spawn_events([real], min_date, max_date):
            start = dates.to_utc(
                dates.delete_timezone(item.local_start)
            ).replace(second=0, microsecond=0)
            if start == date:
                return item

        assert False, "lookup for %s failed" % id


class EventOrderIndex(EventIndex):

    def __init__(self, catalog, state, initial_index=None):
        assert state in ('submitted', 'published', 'archived', 'hidden')
        self.state = state
        super(EventOrderIndex, self).__init__(catalog, initial_index)

    @property
    def name(self):
        return 'eventorder-%s' % self.state

    def identity(self, event):
        date = dates.delete_timezone(event.local_start)
        return '%s;%s' % (date.strftime('%y.%m.%d-%H:%M'), event.id)

    def identity_id(self, identity):
        return identity[15:]

    def identity_date(self, identity):
        return dates.to_utc(datetime.strptime(identity[:14], '%y.%m.%d-%H:%M'))

    def event_by_identity(self, identity):

        id = self.identity_id(identity)
        date = self.identity_date(identity)

        return self.event_by_id_and_date(id, date)

    def reindex(self):

        events = self.catalog.query(review_state=self.state)
        self.index = sortedset()

        self.update(events)
        self.generate_metadata()

    def update(self, events):
        if events:
            self.remove(events)

        managed = (e for e in events if e.review_state == self.state)

        for event in self.spawn_events(managed):
            self.index.add(self.identity(event))

    def remove(self, events):
        assert events

        ids = [r.id for r in events]
        stale = set(i for i in self.index if self.identity_id(i) in ids)

        self.index = sortedset(self.index - stale)

    def generate_metadata(self):
        """Creates a metaindex, indexing the date positions by date.
        For example:

        Events:
            0 -> 01.01.2012
            1 -> 05.01.2012
            2 -> 06.01.2012

        Resulting metaindex:
            01.01.2012 -> 0
            02.01.2012 -> 1
            03.01.2012 -> 1
            04.01.2012 -> 1
            05.01.2012 -> 1
            06.01.2012 -> 2

        The idea is for every possible date between the first event
        and the last event to point to an event in the index. This allows
        for fast pagination using slicing.

        """

        if not self.index:
            if self.get_metadata('dateindex'):
                self.set_metadata('dateindex', None)
            return

        dt = lambda i: i if i is None else self.identity_date(i).date()

        keydates = {}

        # set the key positions (first occurrence of each date)
        for ix, date in enumerate(map(dt, self.index)):
            if date not in keydates:
                keydates[date] = ix

        # fill the gaps
        dateindex = {}
        for prev, curr, next in previous_and_next(sorted(keydates.keys())):
            if prev is None:
                continue
            if prev == curr:
                continue

            for date in dates.days_between(prev, curr):
                dateindex[date] = keydates[curr]

        # merge
        dateindex.update(keydates)

        self.set_metadata('dateindex', dateindex)

    def by_range(self, start, end):

        if not start and not end:
            return self.index

        if not self.index:
            return []

        first_date = self.identity_date(self.index[0])
        last_date = self.identity_date(self.index[-1])

        last_date = datetime(
            last_date.year,
            last_date.month,
            last_date.day,
            tzinfo=end.tzinfo
        ) + timedelta(days=1, microseconds=-1)

        if not dates.overlaps(first_date, last_date, start, end):
            return []

        dateindex = self.get_metadata('dateindex')

        # use whatever timezone is given, because a search for a range cannot
        # be normalized, since the day of this search is not really a concept
        # that exists globally, everyone calls a different utc time a day
        start = (start or first_date).date()
        end = (end or last_date).date()

        if start <= first_date.date():
            startrange = 0
        else:
            startrange = dateindex[start]

        if end >= last_date.date():
            endrange = len(self.index)
        else:
            endrange = dateindex[end + timedelta(days=1)]

        return self.index[startrange:endrange]

    def limit_to_subset(self, index, subset):

        if subset is None:
            return index

        ids = set(brain.id for brain in subset)
        return filter(lambda i: self.identity_id(i) in ids, index)

    def lazy_list(self, start, end, subset=None):
        subindex = self.by_range(start, end)
        subindex = self.limit_to_subset(subindex, subset)
        get_item = lambda i: self.event_by_identity(subindex[i])

        return LazyList(get_item, len(subindex))


class EventsDirectoryCatalog(DirectoryCatalog):

    grok.context(IEventsDirectory)
    grok.provides(IDirectoryCatalog)

    _lock = Lock()

    def __init__(self, *args, **kwargs):
        self._daterange = dates.default_daterange
        start, end = getattr(dates.DateRanges(), 'custom')
        self._custom_start, self._custom_end = dates.as_range(start, end)
        self._state = 'published'
        self._import_source = ''

        self.subset = None

        super(EventsDirectoryCatalog, self).__init__(*args, **kwargs)

        self.ix_submitted = self.index_for_state('submitted')
        self.ix_published = self.index_for_state('published')
        self.ix_hidden = self.index_for_state('hidden')
        self.indices = dict(
            submitted=self.ix_submitted, published=self.ix_published,
            hidden=self.ix_hidden
        )

    def daterange_dates(self):
        if self._daterange == 'custom':
            return self._custom_start, self._custom_end
        else:
            return getattr(dates.DateRanges(), self._daterange)

    def index_for_state(self, state):
        return EventOrderIndex(self, state)

    @synchronized(_lock)
    def reindex(self):
        for ix in self.indices.values():
            ix.reindex()

    @property
    def submitted_count(self):
        """ Returns the submitted count depending on the current date filter
        but independent of the current state filter.

        This needs to loop through all elements again so use only if needed.
        """

        results = self.catalog(
            path={'query': self.path, 'depth': 1},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('submitted', )
        )

        submitted_count = 0

        items = []
        for item in results:
            if self.directory.allow_action('publish', item):
                items.append(item)

        for spawn in self.spawn(items):
            submitted_count += 1

        return submitted_count

    def get_daterange(self):
        return self._daterange

    @instance.clearafter
    def set_daterange(self, range):
        assert dates.is_valid_daterange(range), "invalid date range %s" % range
        self._daterange = range

    daterange = property(get_daterange, set_daterange)

    def custom_start_date(self):
        return self._custom_start

    def custom_end_date(self):
        return self._custom_end

    @instance.clearafter
    def set_custom_dates(self, start, end):
        self._custom_start, self._custom_end = dates.as_range(
            start, end
        )

    def get_state(self):
        return self._state

    @instance.clearafter
    def set_state(self, state):
        # as an added security measure it is not yet possible to
        # query for previewed events as they should only be availble
        # to the user with the right token (see form.py)
        assert state in ('submitted', 'published', 'archived', 'hidden')
        self._state = state

    state = property(get_state, set_state)

    def get_import_source(self):
        return self._import_source

    @instance.clearafter
    def set_import_source(self, source):
        if not source in [s['id'] for s in self.import_sources()]:
            source = ''
        self._import_source = source

    import_source = property(get_import_source, set_import_source)

    def sortkey(self):
        return lambda i: i.start

    def query(self, review_state=None, **kwargs):
        review_state = review_state or self.state

        if self.import_source == '':
            results = self.catalog(
                path={'query': self.path, 'depth': 1},
                object_provides=IEventsDirectoryItem.__identifier__,
                review_state=review_state,
                **kwargs
            )
        else:
            path = '/'.join(self.directory.getPhysicalPath()) + '/'
            results = self.catalog(
                path={'query': self.path, 'depth': 1},
                object_provides=IExternalEvent.__identifier__,
                review_state=review_state,
                source=path + self.import_source,
                **kwargs
            )

        return results

    def spawn(self, realitems, start=None, end=None):
        if not all((start, end)):
            start, end = self.daterange_dates()

        for item in realitems:
            for occurrence in recurrence.occurrences(item, start, end):
                for split in recurrence.split_days(occurrence):
                    if dates.overlaps(start, end, split.start, split.end):
                        yield split

    def hide_blocked(self, realitems):
        """ Returns a generator filtering real items, with the submitted items
        which the user can see but not publish left out.

        As this could be reeealy slow we do a bit of a shortcut by only
        checking for the publish action right on the item. The guard is
        in place anyway, so the user will at worst see an item with which
        he can't do anything.

        """

        if self.state != 'submitted':
            return (r for r in realitems)

        key = lambda i: i.review_state != 'submitted' \
            or self.directory.allow_action('publish', i)

        return ifilter(key, realitems)

    @property
    def lazy_list(self):
        start, end = self.daterange_dates()
        self.limit_subset_to_source()
        return self.indices[self.state].lazy_list(start, end, self.subset)

    def filter(self, term):
        nonempty_terms = [t for t in term.values() if t != u'!empty']

        if len(nonempty_terms) == 0:
            return self.items()

        self.subset = super(EventsDirectoryCatalog, self).filter(term)
        return sorted(
            self.spawn(self.hide_blocked(self.subset)), key=self.sortkey()
        )

    def search(self, text):
        self.subset = super(EventsDirectoryCatalog, self).search(text)
        return sorted(
            self.spawn(self.hide_blocked(self.subset)), key=self.sortkey()
        )

    def limit_subset_to_source(self):
        if self.import_source != '' and self.subset == None:
            self.subset = sorted(self.query(), key=self.sortkey())

    def export(self, search=None, term=None, max=None, **kw):
        # Unfortunantely, the search/filter/items/query functions depend all
        # on some view-state variables - we have to find the subset ourselves.
        if search:
            search = search.replace('*', '')
            search = '"{}"*'.format(search)
            kw['SearchableText'] = search

        elif term:
            kw['categories'] = {
                'query': term.values(),
                'operator': 'and'
            }

        subset = self.catalog(
            path={'query': self.path, 'depth': 1},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state='published',
            **kw
        )

        # Unfortunantely, we cannot query for 'not having an interface' nor
        # not having not the attribute 'source' - we have to build the
        # difference of the query above and all external events!
        external = self.catalog(
            path={'query': self.path, 'depth': 1},
            object_provides=IExternalEvent.__identifier__,
            review_state='published',
            **kw
        )
        if len(external):
            subset_ids = list(
                frozenset([brain.id for brain in subset]) -
                frozenset([brain.id for brain in external])
            )
            subset = [brain for brain in subset if brain.id in subset_ids]

        if term:
            filter_key = partial(is_exact_match, term=term)
            subset = filter(filter_key, subset)

        subset = sorted(subset, key=self.sortkey())

        # Get lazy list from indexer using the subset
        start, end = getattr(dates.DateRanges(), 'this_and_next_year')
        ll = self.ix_published.lazy_list(start, end, subset)

        # Check if upper limit is valid
        if not isinstance(max, (int, long)) or (max <= 0):
            max = len(ll)

        return islice(ll, max)

    def calendar(self, search=None, term=None):
        if search:
            items = super(EventsDirectoryCatalog, self).search(search)
        elif term:
            items = super(EventsDirectoryCatalog, self).filter(term)
        else:
            items = super(EventsDirectoryCatalog, self).items()

        return construct_icalendar(self.directory, items)

    def import_sources(self):
        return self.catalog(
            object_provides=IExternalEventSource.__identifier__
        )



from setuptools import setup, find_packages
import os

name = 'seantis.dir.events'
description = (
    "Directory of upcoming Events."
)
version = '1.2.3'


def get_long_description():
    readme = open('README.rst').read()
    history = open(os.path.join('docs', 'HISTORY.rst')).read()
    contributors = open(os.path.join('docs', 'CONTRIBUTORS.rst')).read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, contributors, history))

zug_require = [
    'ftw.contentmenu',
    'izug.basetheme',
]
teamraum_require = [
    'plonetheme.teamraum'
]
tests_require = [
    'collective.betterbrowser>=0.4',
    'collective.testcaselayer',
    'plone.app.testing',
    'mock'
]

setup(name=name, version=version, description=description,
      long_description=get_long_description(),
      classifiers=[
          "Framework :: Plone",
          "Programming Language :: Python",
      ],
      keywords='',
      author='Seantis GmbH',
      author_email='info@seantis.ch',
      url='https://github.com/seantis/seantis.dir.events',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['seantis', 'seantis.dir'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'Plone>=4.3',
          'python-dateutil',
          'plone.app.event[ploneintegration, dexterity]>=1.0.3',
          'plone.event>=1.0',
          'plone.app.dexterity',
          'plone.synchronize',
          'plone.formwidget.recurrence[z3cform]>=1.1',
          'plone.formwidget.datetime[z3cform]>=1.0',
          'collective.autopermission',
          'collective.dexteritytextindexer',
          'collective.z3cform.mapwidget',
          'collective.js.underscore',
          'seantis.dir.base>=1.7',
          'seantis.plonetools>=0.9',
          'zope.proxy',
          'pytz',
          'python-magic',
          'lxml',
          'blist',
          'functools32',
          'isodate',
          'icalendar>=3.5',
      ],
      extras_require=dict(
          zug=zug_require,
          teamraum=teamraum_require,
          tests=tests_require
      ),
      entry_points="""
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )
